import numpy

import onnx_tool
from .graph import ShapeEngine, ValueExpr
import struct


def __write_float2buf(buf, f):
    ba = struct.pack('f', f)
    buf += ba
    return buf


def __write_int2buf(buf, i, bytes=4):
    ba = i.to_bytes(bytes, 'little', signed=True)
    buf += ba
    return buf


def __write_len2buf(buf, src, bytes=4):
    ba = len(src).to_bytes(bytes, 'little', signed=True)
    buf += ba
    return buf


def __write_str2buf(buf, string):
    if not '\0' in string:
        string += '\0'
    barr = bytes(string, 'utf-8')
    buf += barr
    return buf


DTYPE_INT = 0
DTYPE_FLOAT = 1
DTYPE_STR = 2
DTYPE_FP16 = 3
DTYPE_DOUBLE = 4
DTYPE_INT8 = 5
DTYPE_UINT8 = 6
DTYPE_INT16 = 7
DTYPE_UINT16 = 8
DTYPE_INT64 = 9


def __write_data_type(buf, data):
    if isinstance(data, str) or isinstance(data, bytes):
        buf = __write_int2buf(buf, DTYPE_STR)
    if isinstance(data, int):
        buf = __write_int2buf(buf, DTYPE_INT)
    if isinstance(data, float):
        buf = __write_int2buf(buf, DTYPE_FLOAT)
    return buf


def __write_data(buf, data):
    buf = __write_data_type(buf, data)
    if isinstance(data, str):
        buf = __write_str2buf(buf, data)
    if isinstance(data, int):
        buf = __write_int2buf(buf, data)
    if isinstance(data, float):
        buf = __write_float2buf(buf, data)
    return buf


def __write_ndarray(buf, array: numpy.ndarray):
    def write_npdtype(buf, type):
        if type == numpy.float32:
            buf = __write_int2buf(buf, DTYPE_FLOAT)
        elif type == numpy.float16:
            buf = __write_int2buf(buf, DTYPE_FP16)
        elif type == numpy.int32:
            buf = __write_int2buf(buf, DTYPE_INT)
        elif type == numpy.int8:
            buf = __write_int2buf(buf, DTYPE_INT8)
        elif type == numpy.uint8:
            buf = __write_int2buf(buf, DTYPE_UINT8)
        elif type == numpy.int64:
            buf = __write_int2buf(buf, DTYPE_INT64)
        else:
            raise ValueError(f'Unsupported dtype {type}')
        return buf

    buf = write_npdtype(buf, array.dtype)
    buf = __write_len2buf(buf, array.shape)
    for sval in array.shape:
        buf = __write_int2buf(buf, sval)
    buf += array.tobytes()
    return buf


def serialize_shape_engine(engine: ShapeEngine, filepath):
    binfile = open(filepath, 'wb')
    writebuf = bytearray(0)
    writebuf = __write_len2buf(writebuf, engine.variables.keys())
    for name in engine.variables.keys():
        writebuf = __write_str2buf(writebuf, name)
    writebuf = __write_len2buf(writebuf, engine.tensor_desc.keys())
    for name in engine.tensor_desc.keys():
        writebuf = __write_str2buf(writebuf, name)
        desc = engine.tensor_desc[name]
        writebuf = __write_len2buf(writebuf, desc)
        for d in desc:
            writebuf = __write_data(writebuf, d)

    writebuf = __write_len2buf(writebuf, engine.tensor_epxr.keys())

    def serialize_expr(buf, expr: ValueExpr):
        buf = __write_float2buf(buf, float(expr.alpha))
        buf = __write_float2buf(buf, float(expr.beta))
        buf = __write_int2buf(buf, int(expr.factor))
        buf = __write_int2buf(buf, int(expr.truncmode))
        return buf

    for name in engine.tensor_epxr.keys():
        writebuf = __write_str2buf(writebuf, name)
        expr = engine.tensor_epxr[name]
        srcname = expr[0]
        writebuf = __write_str2buf(writebuf, srcname)
        writebuf = serialize_expr(writebuf, expr[1])

    binfile.write(writebuf)
    binfile.close()


def serialize_graph(graph: onnx_tool.Graph, filepath):
    binfile = open(filepath, 'wb')
    writebuf = bytearray(0)
    writebuf = __write_len2buf(writebuf, graph.nodemap.keys())
    for name in graph.nodemap.keys():
        node = graph.nodemap[name]
        writebuf = __write_str2buf(writebuf, name)
        writebuf = __write_str2buf(writebuf, node.op_type)
        writebuf = __write_len2buf(writebuf, node.input)
        for input in node.input:
            writebuf = __write_str2buf(writebuf, input)
        writebuf = __write_len2buf(writebuf, node.output)
        for output in node.output:
            writebuf = __write_str2buf(writebuf, output)

        writebuf = __write_len2buf(writebuf, node.attr.keys())

        def write_attribute(buf, attrval):
            def write_value(buf, val):
                if isinstance(val, str):
                    buf = __write_str2buf(buf, val)
                if isinstance(val, bytes):
                    buf += val + b'\0'
                if isinstance(val, int):
                    buf = __write_int2buf(buf, val)
                if isinstance(val, float):
                    buf = __write_float2buf(buf, val)
                return buf

            if not isinstance(attrval, list):
                attrval = [attrval]
            buf = __write_len2buf(buf, attrval)
            buf = __write_data_type(buf, attrval[0])
            for val in attrval:
                buf = write_value(buf, val)
            return buf

        for key in node.attr.keys():
            writebuf = __write_str2buf(writebuf, key)
            writebuf = write_attribute(writebuf, node.attr[key])
    writebuf = __write_len2buf(writebuf, graph.initials)
    for name in graph.initials:
        writebuf = __write_str2buf(writebuf, name)
        tensor = graph.tensormap[name]
        writebuf = __write_ndarray(writebuf, tensor.numpy)
    binfile.write(writebuf)
    binfile.close()


def serialize_memory_compression(compressed_mem: {}, max_shape: {}, filepath):
    binfile = open(filepath, 'wb')
    writebuf = bytearray(0)
    writebuf = __write_int2buf(writebuf, compressed_mem[1], 8)
    compress_map = compressed_mem[0]
    writebuf = __write_len2buf(writebuf, compress_map.keys(), 8)
    for key in compress_map.keys():
        block = compress_map[key]
        writebuf = __write_str2buf(writebuf, key)
        writebuf = __write_int2buf(writebuf, block[0], 8)
        writebuf = __write_int2buf(writebuf, block[1], 8)
    writebuf = __write_len2buf(writebuf, max_shape.keys(), 8)
    for key in max_shape.keys():
        writebuf = __write_str2buf(writebuf, key)
        writebuf = __write_int2buf(writebuf, max_shape[key], 8)
    binfile.write(writebuf)
    binfile.close()



import copy
import math
import warnings

import numpy
import onnx

from .node import create_node
from .tensor import STATIC_TENSOR, DYNAMIC_TENSOR
from .tensor import get_attribute_data, Tensor, volume
from .utils import VERSION, tuple2str


def __shape_of_initializer__(initial):
    shape = []
    # for nb in tensor.shape.dim
    for nb in initial.dims:
        shape.append(nb)
    return shape


_SHAPE_TENSORS = {
    'Reshape': ('1of2',),
    'Resize': ('2of3', '3of4', '1of2'),
    'Upsample': ('2of3', '3of4', '1of2'),
    'Expand': ('1of2',),
    'Slice': ('1,2of3', '1,2,3of4', '1,2,3,4of5'),
    'ConstantOfShape': ('0of1',),
    'Tile': ('1of2',),
    'Range': ('0,1,2of3',),
    'OneHot': ('1of3',),
    'TopK': ('1of2',),
    'Pad': ('1of2', '1of3',),
    'NonMaxSuppression': ('2of5',),
    'Split': ('1of2',),
    'Unsqueeze': ('1of2',),
    'Squeeze': ('1of2',)
}


def _contains_shape_tensor(n):
    nodeset = _SHAPE_TENSORS.keys()
    shape_tensors = []
    if n.op_type in nodeset:
        tensor_descs = _SHAPE_TENSORS[n.op_type]
        for desc in tensor_descs:
            strs = desc.split('of')
            indice = strs[0]
            count = int(strs[1])
            if len(n.input) == count:
                indistr = indice.split(',')
                for istr in indistr:
                    shape_tensors.append(n.input[int(istr)])
    return shape_tensors


class ValueExpr():
    def __init__(self, srcrange: [], dstrange: []):
        self.alpha = 1
        self.beta = 0
        self.factor = 0
        self.truncmode = 0
        self.build_expr(srcrange, dstrange)

    def error(self, srcrange, dstrange):
        err = 0
        for sval, dval in zip(srcrange, dstrange):
            err += abs(self.__call__(sval) - dval)
        return err

    def build_expr(self, srcrange: [], dstrange: []):
        lastidx = len(srcrange) - 1
        srclen = srcrange[lastidx] - srcrange[0]
        dstlen = dstrange[lastidx] - dstrange[0]
        if srclen <= dstlen:
            self.factor = 0
            self.alpha = round(dstlen / srclen)
            self.beta = round(dstrange[lastidx] - srcrange[lastidx] * self.alpha)
        else:
            self.factor = 1
            self.alpha = round(srclen / dstlen)
            self.beta = round(dstrange[lastidx] - srcrange[lastidx] / self.alpha)

        test = [self.__call__(x) for x in srcrange]
        diff = 0
        for x0, x1 in zip(test, dstrange):
            diff += x0 - x1
        if diff != 0:
            self.truncmode = 1

    def __call__(self, x):
        if self.factor == 0:
            y = self.alpha * x + self.beta
        else:
            y = x / self.alpha + self.beta
        return self.truncate(y)

    def truncate(self, x):
        if self.truncmode == 0:
            return math.ceil(x)
        else:
            return math.floor(x)


class ShapeEngine():
    def __init__(self, input_desc):
        self.input_desc = input_desc
        self.variables = {}
        self.tensor_desc = {}
        self.tensor_epxr = {}
        for key in input_desc.keys():
            self.add_tensor_desc(key, input_desc[key])

    def update_variable(self, key, val):
        self.variables[key] = val

    def __get_shape_from_desc__(self, desc):
        shape = []
        for val in desc:
            if isinstance(val, int):
                shape.append(val)
            else:
                shape.append(self.variables[val])
        return shape

    def add_tensor_desc(self, tensor, desc):
        self.tensor_desc[tensor] = desc

    def update_tensor_desc(self, tensor, axis, vkey):
        desc = self.tensor_desc[tensor]
        assert (not isinstance(desc[axis], int))
        desc[axis] = vkey

    def get_tensor_desc(self, tensor):
        if tensor not in self.tensor_desc.keys():
            return None
        return self.tensor_desc[tensor]

    def add_expr(self, var_name, src_name, expr):
        self.tensor_epxr[var_name] = [src_name, expr]
        self.variables[var_name] = 0

    def update_variables(self):
        for key in self.tensor_epxr.keys():
            expr = self.tensor_epxr[key]
            self.variables[key] = expr[1](self.variables[expr[0]])

    def get_tensorshape(self, tname):
        desc = self.tensor_desc[tname]
        shape = self.__get_shape_from_desc__(desc)
        return shape

    def generate_input(self):
        tmp_input = {}
        for key in self.input_desc:
            shape = self.get_tensorshape(key)
            tmp_input[key] = numpy.zeros(shape)
        return tmp_input


class Graph():
    def __init__(self, g: onnx.GraphProto, constant_folding: bool = True, noderename: bool = False, verbose=False):
        self.verbose = verbose
        self.nodemap = {}
        self.tensormap = {}
        self.producedby = {}
        self.consumedby = {}
        self.rawgraph = g
        self.initials = []
        self.dynamics = []
        self.input = []
        self.output = []
        self.valid_shape = False
        self.valid_profile = False

        if g is not None:
            self.__init_graph_from_onnxproto__(g, noderename)
            self.__constant_search__(constant_folding)
            self.__update_nodes_tensors__(constant_folding)
            self.__find_shape_tensors__()

    def log(self, str):
        if self.verbose:
            print(str)

    def __update_nodes_tensors__(self, constant_folding):
        from .utils import timer
        tm = timer()
        if constant_folding:
            rmlist = []
            for key in self.nodemap:
                if self.nodemap[key].constant:
                    rmlist.append(key)
            self.log(f'GraphProto Nodes Count:{len(self.rawgraph.node)}')
            self.log(f'Contant folding {rmlist[:3]}... {len(rmlist)} Nodes')
            for key in rmlist:
                self.nodemap.pop(key)

        self.initials = []
        self.dynamics = []

        if constant_folding:
            for name in self.nodemap.keys():
                for tname in self.nodemap[name].input:
                    if self.tensormap[tname].type == STATIC_TENSOR:
                        if tname not in self.initials:
                            self.initials.append(tname)
                    if self.tensormap[tname].type == DYNAMIC_TENSOR:
                        if tname not in self.dynamics:
                            self.dynamics.append(tname)
                for tname in self.nodemap[name].output:
                    if tname not in self.dynamics:
                        self.dynamics.append(tname)
        else:
            for initial in self.rawgraph.initializer:
                self.initials.append(initial.name)
            self.dynamics.extend(self.input)
            for name in self.nodemap.keys():
                for tname in self.nodemap[name].input:
                    if tname not in self.initials and tname not in self.dynamics:
                        self.dynamics.append(tname)
                for tname in self.nodemap[name].output:
                    if tname not in self.dynamics:
                        self.dynamics.append(tname)

        valid_tensors = []
        valid_tensors.extend(self.initials)
        valid_tensors.extend(self.dynamics)
        rm_list = []
        for tname in self.tensormap.keys():
            if tname not in valid_tensors:
                rm_list.append(tname)
        for tname in rm_list:
            self.tensormap.pop(tname)

        self.input = []
        self.output = []
        for name in self.nodemap.keys():
            node = self.nodemap[name]
            for tensor in node.input:
                if tensor not in self.producedby and tensor in self.dynamics:
                    if tensor not in self.input:
                        self.input.append(tensor)
            for tensor in node.output:
                if tensor not in self.consumedby or len(self.consumedby[tensor]) == 0:
                    if tensor not in self.output:
                        self.output.append(tensor)
        self.__update_consumer_producer__()
        self.log(f'Update Nodes Tensors  Time Elapsed {tm.stop()}')

    def __is_node_constant__(self, node):
        constant_node = True
        if node.op_type in ['DequantizeLinear']:
            return False
        for tname in node.input:
            if self.tensormap[tname].type == DYNAMIC_TENSOR:
                constant_node = False
                break
        return constant_node

    def __constant_search__(self, constant_folding):
        from .utils import timer
        tm = timer()
        for name in self.nodemap.keys():
            node = self.nodemap[name]
            if hasattr(node, 'constant'):
                continue
            constant_node = self.__is_node_constant__(node)
            node.constant = constant_node
            if constant_node:
                search_nodes = [name]
                while len(search_nodes) > 0:
                    this_node = self.nodemap[search_nodes[0]]
                    search_nodes.pop(0)
                    if constant_folding:
                        itensors = []
                        for input in this_node.input:
                            itensors.append(self.tensormap[input])
                            if self.tensormap[input].numpy is None:
                                warnings.warn(f'Tensor {input} has shape only, {name} may has wrong value infer result')
                        otensors = []
                        for output in this_node.output:
                            otensors.append(self.tensormap[output])
                        this_node.value_infer(itensors, otensors)
                        if len(otensors) > 0:
                            for i, output in enumerate(this_node.output):
                                self.tensormap[output].type = STATIC_TENSOR
                    else:
                        for i, output in enumerate(this_node.output):
                            self.tensormap[output].type = STATIC_TENSOR
                    for output in this_node.output:
                        for consumer in self.consumedby[output]:
                            cnode = self.nodemap[consumer]
                            if self.__is_node_constant__(cnode):
                                cnode.constant = True
                                search_nodes.append(consumer)
        self.log(f'Constant Search Time Elapsed {tm.stop()}')

    def __update_consumer_producer__(self):
        self.producedby = {}
        self.consumedby = {}
        for name in self.nodemap:
            node = self.nodemap[name]
            for tensor in node.input:
                if tensor not in self.consumedby:
                    self.consumedby[tensor] = []
                self.consumedby[tensor].append(node.name)
            for tensor in node.output:
                if tensor not in self.producedby:
                    self.producedby[tensor] = []
                self.producedby[tensor].append(node.name)

    def update_tensor_relations(self):
        self.__update_consumer_producer__()
        self.dynamics = list(self.producedby.keys())

    def __init_graph_from_onnxproto__(self, g, noderename):
        if g is None:
            return
        ncount = 0
        from .utils import timer

        tm = timer()
        tm.start()
        for node in g.node:
            newnode = create_node(node)
            if noderename or len(newnode.name) == 0:
                newnode.name = newnode.op_type + '_' + str(ncount)
            ncount += 1
            for tensor in node.input:
                if tensor in self.producedby:
                    for producer in self.producedby[tensor]:
                        newnode.prevnodes.append(self.nodemap[producer])
                if tensor not in self.consumedby:
                    self.consumedby[tensor] = []
                self.consumedby[tensor].append(newnode.name)
                newnode.input.append(tensor)
            for tensor in node.output:
                if tensor not in self.producedby:
                    self.producedby[tensor] = []
                self.producedby[tensor].append(newnode.name)
                newnode.output.append(tensor)

            self.nodemap[newnode.name] = newnode

        self.log(f'Node Init Time Elapsed {tm.stop()}')

        tm.start()
        # init initials first
        for initial in g.initializer:
            tensor = Tensor(initial)
            self.tensormap[initial.name] = tensor

        for input in g.input:
            if input.name not in self.tensormap.keys():
                self.tensormap[input.name] = Tensor(input)

        for output in g.output:
            if output.name not in self.tensormap.keys():
                self.tensormap[output.name] = Tensor(output)

        # init dynamic tensor info
        for valinfo in g.value_info:
            if valinfo.name not in self.tensormap.keys():
                tensor = Tensor(valinfo)
                self.tensormap[valinfo.name] = tensor
        self.log(f'Tensor Init Time Elapsed {tm.stop()}')

        tm.start()
        # dynamic tensor info
        for key in self.nodemap.keys():
            node = self.nodemap[key]
            for tensor in node.input:
                if tensor not in self.tensormap.keys():
                    self.tensormap[tensor] = Tensor(tensor)
            for tensor in node.output:
                if tensor not in self.tensormap.keys():
                    self.tensormap[tensor] = Tensor(tensor)
                if tensor in self.consumedby:
                    for consumer in self.consumedby[tensor]:
                        self.nodemap[node.name].nextnodes.append(self.nodemap[consumer])
        self.log(f'IO Tensor Init Time Elapsed {tm.stop()}')

        self.sparse_model = False
        for key in self.tensormap.keys():
            tensor = self.tensormap[key]
            if tensor.sparsity is not None and tensor.sparsity['ratio'] > 0.4:
                self.sparse_model = True
                break

    def __find_shape_tensors__(self):
        self.shape_tensors = []
        for n in self.nodemap.keys():
            shape_tensors = _contains_shape_tensor(self.nodemap[n])
            for st in shape_tensors:
                self.shape_tensors.append(st)
        self.shape_tensors = set(self.shape_tensors)
        # print(self.shape_tensors)
        for tensor in self.shape_tensors:
            if tensor not in self.initials and tensor in self.producedby.keys():
                searchnodes = self.producedby[tensor]
                while len(searchnodes) > 0:
                    nextnodes = []
                    for nname in searchnodes:
                        node = self.nodemap[nname]
                        node.shape_calc = True
                        if node.op_type == 'Shape':
                            continue
                        for input in node.input:
                            if input not in self.initials and input in self.producedby.keys():
                                producers = self.producedby[input]
                                nextnodes.extend([p for p in producers if self.nodemap[p].shape_calc == False])
                    searchnodes = nextnodes

    def __get_subnodes_byio__(self, inputs: [], outputs: []):
        graph_level0 = []
        graph_level1 = []
        graph_level2 = []
        searchlist = outputs
        while len(searchlist):
            newlist = []
            for name in searchlist:
                if name in self.consumedby:
                    for consumer in self.consumedby[name]:
                        if consumer in graph_level2:
                            continue
                        graph_level2.append(consumer)
                        for tensor in self.nodemap[consumer].output:
                            newlist.append(tensor)
            searchlist = newlist

        searchlist = inputs
        while len(searchlist):
            newlist = []
            for name in searchlist:
                if name in self.consumedby:
                    for consumer in self.consumedby[name]:
                        if consumer in graph_level2 or consumer in graph_level1:
                            continue
                        graph_level1.append(consumer)
                        for tensor in self.nodemap[consumer].output:
                            newlist.append(tensor)
            searchlist = newlist

        for node in self.nodemap.keys():
            if node not in graph_level1 and node not in graph_level2:
                graph_level0.append(node)

        return graph_level0, graph_level1, graph_level2

    def add_initial(self, name, data):
        from .tensor import create_initial_Tensor
        self.initials.append(name)
        if isinstance(data, numpy.ndarray):
            self.tensormap[name] = create_initial_Tensor(name, data)
        else:
            raise NotImplementedError('unsupported data type')

    def get_subgraph(self, inputs: [], outputs: []):
        graph_level0, graph_level1, graph_level2 = self.__get_subnodes_byio__(inputs, outputs)

        graph_level0 = Graph(self.get_onnxgraph_by_nodenames(graph_level0))
        graph_level1 = Graph(self.get_onnxgraph_by_nodenames(graph_level1))
        graph_level2 = Graph(self.get_onnxgraph_by_nodenames(graph_level2))

        group_outputs = [graph_level0.output, graph_level1.output, graph_level2.output]
        group_inputs = [graph_level0.input, graph_level1.input, graph_level2.input]

        extern_outputs = []
        extern_inputs = []
        for ele in group_outputs:
            extern_outputs.extend(ele)
        extern_outputs = set(extern_outputs)

        for ele in group_inputs:
            extern_inputs.extend(ele)
        extern_inputs = set(extern_inputs)

        for inputs in group_inputs:
            extern_outputs = extern_outputs - set(inputs)

        for outputs in group_outputs:
            extern_inputs = extern_inputs - set(outputs)

        if len(extern_inputs) != len(self.input):
            warnings.warn("subgraph input and output tensors can not reverse to raw graph.")

        if len(extern_outputs) != len(self.output):
            warnings.warn("subgraph input and output tensors can not reverse to raw graph.")

        return graph_level0, graph_level1, graph_level2

    def get_initials_from_nodenames(self, nodenames):
        initializer = []
        enqueued = []
        for name in nodenames:
            for input in self.nodemap[name].input:
                if input in self.initials and input not in enqueued:
                    initializer.append(self.tensormap[input].make_tensor_proto())
                    enqueued.append(input)
        return initializer

    def remove_node(self, nodename):
        node = self.nodemap[nodename]
        # update producer
        for o in node.output:
            if o in self.producedby.keys():
                self.producedby[o].remove(nodename)
                if len(self.producedby[o]) == 0:
                    self.producedby.pop(o)
        # update consumer
        for i in node.input:
            if i in self.consumedby.keys():
                self.consumedby[i].remove(nodename)
        self.nodemap.pop(nodename)

    def skip_node(self, nodename):
        node = self.nodemap[nodename]
        count = 0
        for input in node.input:
            if input in self.dynamics:
                count += 1
                indtensor = input
        if count == 1 and len(node.output) == 1:
            self.consumedby[indtensor].remove(nodename)
            if node.output[0] in self.consumedby:
                for con in self.consumedby[node.output[0]]:
                    con_node = self.nodemap[con]
                    for i in range(len(con_node.input)):
                        if con_node.input[i] == node.output[0]:
                            con_node.input[i] = indtensor
                            self.consumedby[indtensor].append(con)
                            break
            else:
                for pro in self.producedby[indtensor]:
                    pro_node = self.nodemap[pro]
                    assert (len(pro_node.output) == 1)
                    pro_node.output[0] = node.output[0]
        self.nodemap.pop(nodename)

    def fuse_subgraph_node_names(self, nodes: [str], nodeop: str, nodename: str, keep_attr=True):
        _inputs, _outputs = self.get_iotensors(nodes, remove_initials=False)
        newnode = onnx.helper.make_node(nodeop, _inputs, _outputs, name=nodename)
        count = 0
        if keep_attr:
            for node in nodes:
                for attribute in self.nodemap[node].proto.attribute:
                    attr = onnx.helper.make_attribute(
                        self.nodemap[node].proto.op_type + str(count) + '_' + attribute.name,
                        get_attribute_data(attribute))
                    newnode.attribute.append(attr)
                count += 1
        from .node import Node
        for name in nodes:
            self.remove_node(name)
        newnode = Node(newnode)
        newnode.input = _inputs
        newnode.output = _outputs
        for i in _inputs:
            self.consumedby[i].append(nodename)
            if i in self.producedby.keys():
                newnode.prevnodes.append(self.producedby[i])
        for o in _outputs:
            self.producedby[o] = [nodename]
            if o in self.consumedby.keys():
                newnode.nextnodes.append(self.consumedby[o])
        self.nodemap[nodename] = newnode

    def fuse_postop_node_names(self, nodes: [str], append_attr: bool, **extrakeys):
        _inputs, _outputs = self.get_iotensors(nodes, remove_initials=False)
        mainnode = self.nodemap[nodes[0]]
        newnode = onnx.helper.make_node(mainnode.op_type, _inputs, _outputs, name=mainnode.name)
        count = 0
        for attr in mainnode.proto.attribute:
            if attr.name == 'postop_count':
                count = onnx.helper.get_attribute_value(attr)
            else:
                newnode.attribute.append(attr)

        for key in extrakeys:
            attr = onnx.helper.make_attribute(key, extrakeys[key])
            newnode.attribute.append(attr)

        if append_attr:
            for nname in nodes:
                if nname == mainnode.name:
                    continue
                pnode = self.nodemap[nname]
                attr = onnx.helper.make_attribute(
                    'postop_' + str(count), pnode.op_type)
                newnode.attribute.append(attr)
                for attribute in pnode.proto.attribute:
                    attr = onnx.helper.make_attribute(
                        'postop_' + str(count) + '_' + attribute.name,
                        get_attribute_data(attribute))
                    newnode.attribute.append(attr)
                count += 1

        attr = onnx.helper.make_attribute('postop_count', count)
        newnode.attribute.append(attr)

        for name in nodes:
            self.remove_node(name)
        newnode = create_node(newnode)
        newnode.input = _inputs
        newnode.output = _outputs
        for i in _inputs:
            if i in self.consumedby.keys():
                self.consumedby[i].append(mainnode.name)
            if i in self.producedby.keys():
                newnode.prevnodes.append(self.producedby[i])
        for o in _outputs:
            self.producedby[o] = [mainnode.name]
            if o in self.consumedby.keys():
                newnode.nextnodes.append(self.consumedby[o])
        self.nodemap[mainnode.name] = newnode

    def fuse_subgraph_iotensors(self, inputs: [], outputs: [], nodeop: str, name_prefix: str = None, keep_attr=True):
        _, nodes, _ = self.__get_subnodes_byio__(inputs, outputs)
        from .fusion import create_descs_from_nodenames, FusionPattern
        descs = create_descs_from_nodenames(self, nodes)
        pattern = FusionPattern(descs)
        nodesls = pattern.search_pattern(self)
        for i, nodes in enumerate(nodesls):
            name = name_prefix + '_' + str(i) if name_prefix is not None else nodes[0]
            self.fuse_subgraph_node_names(nodes, nodeop, name, keep_attr)

    def get_onnxgraph_by_nodenames(self, nodenames):
        if len(nodenames):
            _inputs0, _outputs0 = self.get_iotensors(nodenames)
            graph_level0 = self.topsort_nodes(nodenames, _inputs0)
            subgraph = self.make_graph_onnx(graph_level0, 'subgraph', _inputs0, _outputs0)
            return subgraph
        return None

    def save_model(self, f: str, shape_only: bool = False, no_shape: bool = False, rawmodel: onnx.ModelProto = None):
        if len(self.nodemap.keys()) == 0:
            warnings.warn(f'Empty graph {f} to save')
            return
        graph = self.make_graph_onnx(self.nodemap.keys(), 'graph', self.input, self.output,
                                     with_initializer=not shape_only, with_shape_info=not no_shape)
        if graph is not None and f is not None:
            attr = {}
            attr['producer_name'] = 'onnx_tool'
            attr['producer_version'] = 'v' + VERSION
            model = onnx.helper.make_model(graph, **attr)
            if rawmodel is not None:
                model.ir_version = rawmodel.ir_version
                model.opset_import[0].version = rawmodel.opset_import[0].version
            onnx.save_model(model, f)

    def make_graph_onnx(self, nodenames, gname, inputnames, outputnames, with_initializer=True, with_shape_info=True):
        nodes = []
        for name in nodenames:
            nodes.append(self.nodemap[name].make_nodeproto())

        initializer = None
        if with_initializer:
            initializer = self.get_initials_from_nodenames(nodenames)

        inputs = []
        outputs = []
        for name in inputnames:
            if name in self.tensormap:
                proto = self.tensormap[name].make_value_proto(make_dummy=True)
                if proto is not None:
                    inputs.append(proto)
            else:
                inputs.append(onnx.helper.make_tensor_value_info(name, 1, None))
        for name in outputnames:
            if name in self.tensormap:
                proto = self.tensormap[name].make_value_proto(make_dummy=True)
                if proto is not None:
                    outputs.append(proto)
        value_infos = []
        if with_shape_info:
            for key in self.dynamics:
                if key in self.input or key in self.output:
                    continue
                tensor = self.tensormap[key]
                vinfo = tensor.make_value_proto()
                if vinfo is None:
                    continue
                value_infos.append(vinfo)
            if not with_initializer:
                for key in self.initials:
                    tensor = self.tensormap[key]
                    vinfo = tensor.make_value_proto()
                    if vinfo is None:
                        continue
                    value_infos.append(vinfo)
        graph = onnx.helper.make_graph(nodes=nodes, name=gname, inputs=inputs, outputs=outputs, initializer=initializer,
                                       value_info=value_infos)
        return graph

    def topsort_nodes(self, node_names, input_names):
        # update
        produced_by = {}
        for name in node_names:
            node = self.nodemap[name]
            for tname in node.output:
                produced_by[tname] = name

        consumed_by = {}
        dependencies = {}
        for name in node_names:
            node = self.nodemap[name]
            count = 0
            for tname in node.input:
                if tname in produced_by.keys():
                    count += 1
                    if tname in consumed_by.keys():
                        consumed_by[tname].append(name)
                    else:
                        consumed_by[tname] = [name]
            dependencies[name] = count

        ordered_nodes = []
        queue = []
        while True:
            for name in node_names:
                if dependencies[name] == 0:
                    queue.append(name)
                    ordered_nodes.append(name)
                    dependencies[name] -= 1
            if len(queue) == 0:
                break
            for name in queue:
                node = self.nodemap[name]
                for o in node.output:
                    if o in consumed_by.keys():
                        for con in consumed_by[o]:
                            dependencies[con] -= 1
            queue.clear()
        return ordered_nodes

    def graph_reorder_nodes(self):
        ordered_nodes = self.topsort_nodes(self.nodemap.keys(), self.input)
        new_map = {}
        for nname in ordered_nodes:
            new_map[nname] = self.nodemap[nname]
        self.nodemap = new_map
        self.update_tensor_relations()

    def get_iotensors(self, nodenames, remove_initials=True):
        intensors = []
        outtensors = []
        for name in nodenames:
            for input in self.nodemap[name].input:
                if len(input) == 0:
                    continue
                if remove_initials and input in self.initials:
                    continue
                if input in self.producedby:
                    producers = self.producedby[input]
                    inner = True
                    for producer in producers:
                        if producer not in nodenames:
                            inner = False
                            break
                    if inner:
                        continue
                if input not in intensors:
                    intensors.append(input)

            for output in self.nodemap[name].output:
                if remove_initials and output in self.initials:
                    continue
                if output in self.consumedby:
                    consumers = self.consumedby[output]
                    inner = True
                    for consumer in consumers:
                        if consumer not in nodenames:
                            inner = False
                            break
                    if inner:
                        continue
                if output not in outtensors:
                    outtensors.append(output)

        return intensors, outtensors

    def update_input_by_map(self, inputs: {}):
        for key in inputs.keys():
            if key in self.tensormap.keys():
                self.tensormap[key].update_tensor(inputs[key])

    def get_dynamic_tensors(self):
        dtensors = {}
        import copy
        for key in self.dynamics:
            dtensors[key] = copy.deepcopy(self.tensormap[key])
        return dtensors

    def check_inputs(self):
        for name in self.input:
            shape = self.tensormap[name].shape
            for val in shape:
                if isinstance(val, str):
                    return False, name
                if val < 0:
                    return False, name
        return True, None

    def shape_infer(self, inputs: {} = None):
        self.valid_shape = False
        if inputs is not None:
            self.update_input_by_map(inputs)
        in_valid, tname = self.check_inputs()
        if not in_valid:
            raise ValueError(
                f"The input tensor {tname}'s shape {self.tensormap[tname].shape2str()} is not valid, Please set it to a valid shape.")
        self.shapeinfer_optime_map = {}
        from .utils import timer
        tm = timer()
        for key in self.nodemap.keys():
            tm.start()
            node = self.nodemap[key]
            itensors = []
            for input in node.input:
                itensors.append(self.tensormap[input])
            otensors = []
            for output in node.output:
                otensors.append(self.tensormap[output])

            if node.shape_calc:
                node.value_infer(itensors, otensors)
            else:
                node.shape_infer(itensors, otensors)

            if node.op_type in self.shapeinfer_optime_map.keys():
                self.shapeinfer_optime_map[node.op_type] += tm.stop()
            else:
                self.shapeinfer_optime_map[node.op_type] = tm.stop()
        self.log(self.shapeinfer_optime_map)
        self.valid_shape = True

    def value_infer(self, inputs: {}):
        self.update_input_by_map(inputs)
        for key in self.nodemap.keys():
            node = self.nodemap[key]
            itensors = []
            for input in node.input:
                itensors.append(self.tensormap[input])
            otensors = []
            for output in node.output:
                otensors.append(self.tensormap[output])
            node.value_infer(itensors, otensors)
        outputs = []
        for output in self.output:
            outputs.append(self.tensormap[output].numpy)
        return outputs

    def add_dump_tensors(self, dump_tensor_names: []):
        for name in dump_tensor_names:
            if name in self.tensormap.keys():
                if name not in self.output:
                    self.output.append(name)

    def shape_regress(self, input_desc: {}, input_range: {}):
        shapeengine = ShapeEngine(input_desc)

        for key in input_range.keys():
            shapeengine.update_variable(key, input_range[key][1])
        tmp_input = shapeengine.generate_input()
        self.shape_infer(tmp_input)
        maxtensormap = self.get_dynamic_tensors()

        for key in input_range.keys():
            shapeengine.update_variable(key, input_range[key][0])
        tmp_input = shapeengine.generate_input()
        self.shape_infer(tmp_input)
        mintensormap = self.get_dynamic_tensors()

        for key in mintensormap.keys():
            if key in input_desc.keys():
                continue
            shape_desc = []
            minshape = mintensormap[key].shape
            maxshape = maxtensormap[key].shape
            for i, a in zip(minshape, maxshape):
                if i == a:
                    shape_desc.append(i)
                else:
                    shape_desc.append('')
            shapeengine.add_tensor_desc(key, shape_desc)

        vcount = 0
        for key in input_range.keys():
            minv = input_range[key][0]
            maxv = input_range[key][1]
            vranges = [(minv + maxv) // 2, maxv]
            shapes_range = []

            for val in vranges:
                shapeengine.update_variable(key, val)
                tmpinputs = shapeengine.generate_input()
                self.shape_infer(tmpinputs)
                shapes_range.append(self.get_dynamic_tensors())

            shapeengine.update_variable(key, input_range[key][0])
            srcrange = [minv, ] + vranges
            dstranges = [srcrange]
            tensor_range_map = {}

            def range_in_rangelist(range, rangelist):
                for i, r in enumerate(rangelist):
                    flag = True
                    for sv, dv in zip(range, r):
                        if sv != dv:
                            flag = False
                            break
                    if flag:
                        return i
                return -1

            for vkey in mintensormap.keys():
                if vkey in input_desc.keys():
                    continue
                shapes = [mintensormap[vkey].shape, shapes_range[0][vkey].shape, shapes_range[1][vkey].shape]
                for i in range(len(shapes[0])):
                    if shapes[0][i] != shapes[1][i] or shapes[0][i] != shapes[2][i]:
                        newrange = [val[i] for val in shapes]
                        idx = range_in_rangelist(newrange, dstranges)
                        if idx == -1:
                            nodename = self.producedby[vkey][0]
                            srcvalkey = []
                            for iname in self.nodemap[nodename].input:
                                desc = shapeengine.get_tensor_desc(iname)
                                if desc is not None:
                                    for ele in desc:
                                        if not isinstance(ele, int) and ele.isnumeric():
                                            if int(ele) >= vcount:
                                                srcvalkey.append(ele)
                            dstranges.append(newrange)
                            vidx = vcount + len(dstranges) - 1
                            if len(srcvalkey) == 0:
                                expr = ValueExpr(dstranges[0], newrange)
                                err = expr.error(dstranges[0], newrange)
                                srckey = key
                            else:
                                # TODO multiple src variables comparison
                                srckey = srcvalkey[0]
                                expr = ValueExpr(tensor_range_map[srckey], newrange)
                                err = expr.error(tensor_range_map[srckey], newrange)
                            if err > 0:
                                warnings.warn(f'src key {srckey} dst key {vidx} error:{err}')

                            shapeengine.add_expr(str(vidx), srckey, expr)
                            tensor_range_map[str(vidx)] = newrange
                            valkey = str(vidx)
                        else:
                            if idx == 0:
                                valkey = key
                            else:
                                vidx = vcount + idx
                                valkey = str(vidx)
                        shapeengine.update_tensor_desc(vkey, i, valkey)
            vcount += len(dstranges)
        return shapeengine

    def compress_memory(self, size_padding=64):
        tensor_in_mem = copy.deepcopy(self.input)
        tensor_mem_per_node = []
        tensor_consumed = {}
        for node in self.nodemap.keys():
            node_ = self.nodemap[node]
            for name in node_.output:
                tensor_in_mem.append(name)
            new_list = copy.deepcopy(tensor_in_mem)
            tensor_mem_per_node.append(new_list)
            for name in node_.input:
                if name in self.consumedby:
                    if name in tensor_consumed:
                        tensor_consumed[name].append(node)
                    else:
                        tensor_consumed[name] = [node]
                    consumers = self.consumedby[name]
                    if len(tensor_consumed[name]) == len(consumers):
                        if name in tensor_in_mem:
                            tensor_in_mem.remove(name)
        for output in self.output:
            if output not in tensor_in_mem:
                warnings.warn(f'tensor list is wrong, {output} is missing!')
        # print(tensor_mem_per_node)
        compress_mem = {}
        mem_tags = []
        mem_block = []
        for nodetensors in tensor_mem_per_node:
            # clear mem tags
            for i, tag in enumerate(mem_tags):
                if tag == "":
                    continue
                if tag not in nodetensors:
                    premerge = False
                    if i >= 1 and mem_tags[i - 1] == "":
                        premerge = True
                    nextmerge = False
                    if i < len(mem_tags) - 1 and mem_tags[i + 1] == "":
                        nextmerge = True
                    if premerge and nextmerge:
                        newblock = [mem_block[i - 1][0], mem_block[i + 1][0] + mem_block[i + 1][1]
                                    - mem_block[i - 1][0]]
                        mem_tags.pop(i)
                        mem_tags.pop(i)
                        mem_block.pop(i)
                        mem_block.pop(i)
                        mem_tags[i - 1] = ""
                        mem_block[i - 1] = newblock
                    elif premerge:
                        newblock = [mem_block[i - 1][0], mem_block[i][0] + mem_block[i][1]
                                    - mem_block[i - 1][0]]
                        mem_tags.pop(i)
                        mem_block.pop(i)
                        mem_tags[i - 1] = ""
                        mem_block[i - 1] = newblock
                    elif nextmerge:
                        newblock = [mem_block[i][0], mem_block[i + 1][0] + mem_block[i + 1][1]
                                    - mem_block[i][0]]
                        mem_tags.pop(i)
                        mem_block.pop(i)
                        mem_tags[i] = ""
                        mem_block[i] = newblock
                    else:
                        mem_tags[i] = ""

            # push tensor mem to block
            for tname in nodetensors:
                t_ = self.tensormap[tname]
                size_ = t_.get_memsize()
                size_ = int((size_ + size_padding - 1) // size_padding * size_padding)
                if tname in mem_tags:
                    continue
                block_found = False
                for i, tag in enumerate(mem_tags):
                    if tag == "":
                        if mem_block[i][1] >= size_:
                            remain_size = mem_block[i][1] - size_
                            if remain_size > 1024 * 1024:
                                remain_block = [mem_block[i][0] + size_, remain_size]
                                mem_tags[i] = tname
                                mem_block[i][1] = size_
                                mem_tags.insert(i + 1, "")
                                mem_block.insert(i + 1, remain_block)
                            else:
                                mem_tags[i] = tname
                            block_found = True
                            break
                if not block_found:
                    lastidx = len(mem_tags) - 1
                    if lastidx >= 0 and mem_tags[lastidx] == "":
                        mem_block[lastidx][1] = size_
                        mem_tags[lastidx] = tname
                    else:
                        mem_tags.append(tname)
                        if lastidx >= 0:
                            mem_block.append([mem_block[lastidx][0] + mem_block[lastidx][1], size_])
                        else:
                            mem_block.append([0, size_])

                idx = mem_tags.index(tname)
                compress_mem[tname] = mem_block[idx]

        lastblock = mem_block[-1]
        compress_size = lastblock[0] + lastblock[1]
        # print(compress_mem)
        # validate memory
        for nodetensors in tensor_mem_per_node:
            maxmem = 0
            overlap = False
            for tname in nodetensors:
                block = compress_mem[tname]
                if block[0] + block[1] > maxmem:
                    maxmem = block[0] + block[1]
            for tname in nodetensors:
                block0 = compress_mem[tname]
                if block0[1] == 0:
                    continue
                for tname1 in nodetensors:
                    if tname1 == tname:
                        continue
                    block1 = compress_mem[tname1]
                    if block1[1] == 0:
                        continue
                    if block0[0] >= block1[0] and block0[0] < block1[0] + block1[1]:
                        overlap = True
                        laptensor = [tname, tname1]
                        break
                    if block1[0] >= block0[0] and block1[0] < block0[0] + block0[1]:
                        overlap = True
                        laptensor = [tname, tname1]
                        break
            if maxmem > compress_size:
                warnings.warn(f'Wrong compress total memory size:{compress_size}. larger size detected:{maxmem}')
            if overlap:
                if laptensor[0] != '' and laptensor[
                    1] != '':  # some chaos models use this empty name as an empty tensor name
                    warnings.warn(f'Memory overlap detected!{laptensor[0]} v.s. {laptensor[1]}')

        raw_memsize = 0
        for tname in self.dynamics:
            if tname in self.input or tname in self.output:
                continue
            raw_memsize += self.tensormap[tname].get_memsize()

        self.log(f"Raw memory size (without parameter memory): {raw_memsize:,} bytes")
        self.log(f"Compressed memory size: {compress_size:,} bytes")
        self.log(f'Comression ratio: {compress_size / raw_memsize * 100:.3f}%')
        return compress_mem, compress_size

    def get_compute_graph(self):
        cg = copy.copy(self)
        nodes = []
        rmnodes = []
        for key in cg.nodemap.keys():
            node = cg.nodemap[key]
            if node.shape_calc:
                rmnodes.append(key)
                continue
            dummy_node = True
            for output in node.output:
                dummy = True
                if output in cg.consumedby.keys():
                    for consumer in cg.consumedby[output]:
                        if not cg.nodemap[consumer].shape_calc:
                            dummy = False
                            break
                else:
                    dummy = False
                dummy_node = dummy_node and dummy
            if dummy_node:
                rmnodes.append(key)
                searchnodes = [key]
                while len(searchnodes) > 0:
                    this_node = self.nodemap[searchnodes[0]]
                    searchnodes.pop(0)
                    for input in this_node.input:
                        if input in self.producedby:
                            if len(self.consumedby[input]) == 1:
                                pnodename = self.producedby[input][0]
                                if pnodename not in rmnodes:
                                    rmnodes.append(pnodename)
                                searchnodes.append(pnodename)
                continue
            nodes.append(node.name)
        self.log(f'Total Nodes:{len(cg.nodemap)} Removed Shape Nodes:{len(rmnodes)}')
        for key in rmnodes:
            cg.remove_node(key)
        self.log(f'Compute Graph Nodes:{len(cg.nodemap)}')

        cg.dynamics = []
        cg.dynamics.extend(cg.input)
        cg.dynamics.extend(cg.output)
        for name in cg.nodemap.keys():
            for output in cg.nodemap[name].output:
                if name not in cg.dynamics:
                    cg.dynamics.append(output)

        return cg

    def profile(self):
        self.valid_profile = False
        if not self.valid_shape:
            warnings.warn('Please perform a valid shape_infer() before profile().')
            return
        params_flag_map = {}
        for key in self.initials:
            params_flag_map[key] = 0

        self.macs = [0.0, 0.0]
        self.params = 0
        self.memory = 0
        for key in self.nodemap.keys():
            node = self.nodemap[key]
            itensors = []
            _params = 0
            _memory = 0
            max_sparsity = 0
            block_sparsity = {'blocksize': (1, 1), 'blockratio': 0, 'ratio': 0}
            for input in node.input:
                tensor = self.tensormap[input]
                itensors.append(tensor)
                if input in self.initials:
                    if params_flag_map[input] == 0:
                        elesize = volume(self.tensormap[input].get_shape())
                        _params += elesize
                        _memory += elesize * self.tensormap[input].get_elementsize()
                    params_flag_map[input] += 1
                if tensor.sparsity is not None and tensor.sparsity['ratio'] > max_sparsity:
                    max_sparsity = tensor.sparsity['ratio']
                    block_sparsity = tensor.sparsity
            otensors = []
            for output in node.output:
                otensors.append(self.tensormap[output])
                if node.op_type == 'Constant':
                    # Constant's output tensors are already counted as weight tensors
                    continue
                _memory += self.tensormap[output].get_memsize()
            macs = node.profile(itensors, otensors)
            outshape = (0,)
            if len(node.output) > 0:
                outshape = self.tensormap[node.output[0]].get_shape()
                outshape = (0,) if len(outshape) == 0 else outshape
            inshape = (0,)
            if len(node.input) > 0:
                inshape = self.tensormap[node.input[0]].get_shape()
                inshape = (0,) if len(inshape) == 0 else inshape

            node.macs = macs
            node.inshape = inshape
            node.outshape = outshape
            node.params = _params
            node.memory = _memory
            node.sparsity = block_sparsity
            self.macs[0] += macs[0]
            self.macs[1] += macs[1]
            self.params += _params
            self.memory += _memory

        self.valid_profile = True

    def print_node_map(self, f: str = None, metric='MACs', exclude_ops=None):
        if not self.valid_profile:
            warnings.warn('Please perform a valid profile() before print_node_map().')
            return
        from tabulate import tabulate
        assert (metric in ['MACs', 'FLOPs'])
        print_sparse_table = self.sparse_model
        saveformat = 'txt'
        splitch = 'x'

        if f is not None and '.csv' in f:
            saveformat = 'csv'
            csvformat = True
        else:
            csvformat = False

        ptable = []

        forward_macs = int(round(self.macs[0]))
        backward_macs = int(round(self.macs[1]))
        backward_valid = backward_macs > 0
        params = int(self.params)
        memory = int(self.memory)

        shared_size = 0
        for key in self.tensormap.keys():
            if key in self.initials:
                if key in self.consumedby.keys() and len(self.consumedby[key]) > 1:
                    tensor = self.tensormap[key]
                    shared_size += volume(tensor.get_shape())

        if shared_size > 1024:
            print()
            print('*' * 64)
            print(f'Please note that Weight Tensors Sharing is detected:')
            for key in self.tensormap.keys():
                if key in self.initials:
                    if key in self.consumedby.keys() and len(self.consumedby[key]) > 1:
                        print(f'Tensor:{key} ')
                        print('Shared by: ')
                        for node in self.consumedby[key]:
                            print('           ', node)
                        print()
            print('*' * 64)

        factor = 1
        if metric == 'FLOPs':
            factor = 2

        def num2str(num, csv=False):
            if csv:
                return '{}'.format(num)
            else:
                return '{:,}'.format(num)

        params += 1e-18
        forward_macs += 1e-18
        backward_macs += 1e-18
        for key in self.nodemap.keys():
            node = self.nodemap[key]
            if exclude_ops is not None and node.op_type in exclude_ops:
                continue
            row = [key, self.nodemap[key].op_type]
            if print_sparse_table:
                sparsity = node.sparsity
                row.append(tuple2str(sparsity['blocksize'], splitch))
                row.append('{:.2%}'.format(sparsity['blockratio']))
                row.append('{:.2%}'.format(sparsity['ratio']))
            row.append(num2str(int(node.macs[0]) * factor, csvformat))
            row.append('{:.2%}'.format(node.macs[0] / forward_macs))
            if backward_valid:
                row.append(num2str(int(node.macs[1]) * factor, csvformat))
                row.append('{:.2%}'.format(node.macs[1] / backward_macs))
            row.append(num2str(int(node.memory), csvformat))
            row.append('{:.2%}'.format(node.memory / memory))
            row.append(num2str(int(node.params), csvformat))
            row.append('{:.2%}'.format(node.params / params))
            row.append(tuple2str(node.inshape, splitch))
            row.append(tuple2str(node.outshape, splitch))

            ptable.append(row)
        row = ['Total', '_']
        if print_sparse_table:
            row.append('_')
            row.append('_')
            row.append('_')
        row.append(num2str(int(forward_macs * factor), csvformat))
        row.append('100%')
        if backward_valid:
            row.append(num2str(int(backward_macs * factor), csvformat))
            row.append('100%')
        row.append(num2str(int(memory), csvformat))
        row.append('100%')
        row.append(num2str(int(params), csvformat))
        row.append('100%')
        row.append('_')
        row.append('_')

        ptable.append(row)
        header = ['Name', 'Type']
        if print_sparse_table:
            header.append('Sparse Pattern')
            header.append('Sparse Block Ratio')
            header.append('Sparse Ratio')
        header.extend(
            ['Forward_' + metric, 'FPercent'])
        if backward_valid:
            header.extend(
                ['Backward_' + metric, 'BPercent'])
        header.extend(
            ['Memory', 'MPercent', 'Params',
             'PPercent', 'InShape',
             'OutShape'])

        if f is None:
            print(tabulate(ptable, headers=header))
        else:
            fp = open(f, 'w')
            if saveformat == 'csv':
                headerstr = ''
                for i, item in enumerate(header):
                    headerstr += item
                    if i < len(header) - 1:
                        headerstr += ','
                headerstr += '\n'
                fp.write(headerstr)
                for row in ptable:
                    str = ''
                    for i, ele in enumerate(row):
                        str += ele
                        if i != len(row) - 1:
                            str += ','
                    str += '\n'
                    fp.write(str)
            else:
                fp.write(tabulate(ptable, headers=header))
            fp.close()



import warnings

import numpy

import onnx_tool

'''
f32 layout: OCxIC ONNX default, len(shape)==2 2-d weight only

block: 0 per tensor
       -1 per oc
       >0 per block on ic
'''


def find_min_max(f32: numpy.ndarray, block):
    if block == 0:
        return numpy.array(f32.min(keepdims=False)), numpy.array(f32.max(keepdims=False))
    if block == -1:
        fmin = [f32[i, :].min() for i in numpy.ndindex(*f32.shape[:-1])]
        fmax = [f32[i, :].max() for i in numpy.ndindex(*f32.shape[:-1])]
        return numpy.array(fmin), numpy.array(fmax)
    if block > 0:
        fmin = []
        fmax = []
        for i in range(f32.shape[0]):
            tmin = []
            tmax = []
            for j in range(0, f32.shape[1], block):
                tmin.append(f32[i, j:j + block].min())
                tmax.append(f32[i, j:j + block].max())
            fmin.append(tmin)
            fmax.append(tmax)
        return numpy.array(fmin), numpy.array(fmax)


def get_symmetric_parameter(fmin, fmax, bits):
    alpha = max(abs(fmin), abs(fmax))
    scale = alpha / (2 ** (bits - 1) - 1)
    return scale, 0


def get_asymmetric_u8_parameter(fmin, fmax, bits):
    fmin = min(fmin, 0)
    fmax = max(fmax, 0)
    delta = fmax - fmin
    scale = delta / (2 ** bits - 1)
    zp = round(-fmin / scale)
    return scale, zp


def symmetric_quant(fval, scale, *agrs):
    return numpy.clip(round(fval / scale), -128, 127)


def asymmetric_quant(fval, scale, zp):
    return numpy.clip(round(fval / scale) + zp, 0, 255)


def symmetric_dequant(qval, scale, *agrs):
    return qval * scale


def asymmetric_dequant(qval, scale, zp):
    return (qval - zp) * scale


def quantize(f32: numpy.ndarray, block: int = -1, type: str = 'sym', bits: int = 8):
    fmin, fmax = find_min_max(f32, block)
    if type == 'sym':
        get_param = get_symmetric_parameter
        quant_val = symmetric_quant
    elif type == 'asym':
        get_param = get_asymmetric_u8_parameter
        quant_val = asymmetric_quant
    if len(fmin.shape) == 0:
        Q = numpy.zeros_like(f32, dtype=numpy.int32)
        scale, zp = get_param(fmin, fmax, bits)
        for i in numpy.ndindex(*f32.shape):
            Q[i] = quant_val(f32[i], scale, zp)
        return Q, numpy.array(scale), numpy.array(zp)
    if len(fmin.shape) == 1:
        Q = numpy.zeros_like(f32, dtype=numpy.int32)
        scale = numpy.ones_like(fmin, dtype=numpy.float32)
        zp = numpy.zeros_like(fmin, dtype=numpy.int32)
        for i in range(len(fmin)):
            scale[i], zp[i] = get_param(fmin[i], fmax[i], bits)
        for i in numpy.ndindex(*f32.shape):
            Q[i] = quant_val(f32[i], scale[i[0]], zp[i[0]])
        return Q, scale, zp
    if len(fmin.shape) == 2:
        Q = numpy.zeros_like(f32, dtype=numpy.int32)
        scale = numpy.ones_like(fmin, dtype=numpy.float32)
        zp = numpy.zeros_like(fmin, dtype=numpy.int32)
        for i in numpy.ndindex(*fmin.shape):
            scale[i], zp[i] = get_param(fmin[i], fmax[i], bits)
        for i in numpy.ndindex(*f32.shape):
            blk_idx = i[1] // block
            Q[i] = quant_val(f32[i], scale[i[0], blk_idx], zp[i[0], blk_idx])
        return Q, scale, zp

def pack_4bits(Q_int:numpy.ndarray):
    shape=Q_int.shape
    s4shape=list(shape)
    s4shape[-1]=shape[-1]//2
    s4arr=numpy.ones(s4shape,dtype=numpy.uint8)
    import ctypes
    def cvt_s8bits_to_s4bits(v8):
        sigbit=v8&0x80
        sigbit=sigbit>>4
        v8=v8&0x7
        return v8|sigbit
    for i in numpy.ndindex(*s4shape):
        rawi=list(i)
        rawi[-1]*=2
        rawi1=list(rawi)
        rawi1[-1]+=1
        lo4=cvt_s8bits_to_s4bits(ctypes.c_uint8(Q_int[tuple(rawi)]).value)
        hi4=cvt_s8bits_to_s4bits(ctypes.c_uint8(Q_int[tuple(rawi1)]).value)
        val=lo4|(hi4<<4)
        s4arr[i]=val
    return s4arr

def graph_quantize(g: onnx_tool.Graph, tname: str, block: int = -1, type: str = 'sym', bits: int = 8):
    if tname not in g.initials:
        warnings.warn("Quantize activation tensor is useless")
        return
    quantized_suffix=['_ot_q','_ot_s','_ot_z']
    for suf in quantized_suffix:
        if tname.endswith(suf):
            return
    f32arr = g.tensormap[tname].numpy
    if f32arr.dtype not in [numpy.float32, numpy.float16, numpy.float64]:
        return
    if len(f32arr.shape) != 2:
        return
    if f32arr.dtype is not numpy.float32:
        f32arr = f32arr.astype(numpy.float32)
    name=g.consumedby[tname][0]
    node=g.nodemap[name]
    if isinstance(node,onnx_tool.node.GemmNode):
        if node.transB==0:
            f32arr=f32arr.transpose()
    node.set_attr('OTQ_Block',block)
    node.set_attr('OTQ_Type',type)
    node.set_attr('OTQ_Bits',bits)
    Q, scale, zp = quantize(f32arr, block, type, bits)
    tname_q = tname + '_ot_q'
    tname_s = tname + '_ot_s'
    tname_z = tname + '_ot_z'
    if bits == 8:
        if type == 'asym':
            Q = Q.astype(numpy.uint8)
            zp = zp.astype(numpy.uint8)
        else:
            Q = Q.astype(numpy.int8)
    if bits == 4:
        scale = scale.astype(numpy.float16)
        if type == 'asym':
            Q = pack_4bits(Q)
            zp = zp.astype(numpy.uint8)
        else:
            Q = pack_4bits(Q)

    g.add_initial(tname_q, Q)
    g.add_initial(tname_s, scale)
    if type == 'asym':
        g.add_initial(tname_z, zp)
    for nname in g.consumedby[tname]:
        node = g.nodemap[nname]
        idx = node.input.index(tname)
        node.input.remove(tname)
        if type == 'asym':
            node.input.insert(idx, tname_z)
        node.input.insert(idx, tname_s)
        node.input.insert(idx, tname_q)
        if tname_q not in g.consumedby.keys():
            g.consumedby[tname_q] = nname
            g.consumedby[tname_s] = nname
            g.consumedby[tname_z] = nname
        else:
            g.consumedby[tname_q].append(nname)
            g.consumedby[tname_s].append(nname)
            g.consumedby[tname_z].append(nname)
    g.consumedby.pop(tname)
    g.initials.remove(tname)
    g.tensormap.pop(tname)



import numpy
import onnx
import math
import warnings
from typing import List
from .tensor import get_attribute_data, volume, is_valid_ndarray, create_ndarray_f32, onnxdtype2npdtype, Tensor
from .utils import NODE_REGISTRY

'''
Real MACs: the number of x86 instructions to finish numeric compute.
From a low-level view, float-point multiple and add is much expensive than
bit movement. But current hardware has added more and more float-point
units to accelerate this. So, we should treat every instruction equally to
measure the complexity of the model.


'''

MUL_MACS = 1
ADD_MACS = 1
CMP_MACS = 1
DIV_MACS = 4  # refers to vrcp14ps
# following refers to https://github.com/reyoung/avx_mathfun
EXP_MACS = 32
POW_MACS = EXP_MACS  # similar with exp
LOG_MACS = 43
SIN_MACS = 39
COS_MACS = 39

TANH_MACS = EXP_MACS + 2 * ADD_MACS + DIV_MACS  # (e^2x-1)/(e^2x+1)
ATANH_MACS = LOG_MACS + 2 * ADD_MACS + DIV_MACS + MUL_MACS  # 1/2*ln((1+x)/(1-x))

SQRT_MACS = 24  # refers to vsqrtps
ATAN_MACS = SQRT_MACS + 3 * MUL_MACS + MUL_MACS + DIV_MACS  # refers to "Approximations to inverse tangent function"

RESIZE_LINEAR_MACS = 4
RESIZE_CUBIC_MACS = 8

_SHAPE_TENSORS = {
    'Reshape': ('1of2',),
    'Resize': ('2of3', '3of4'),
    'Slice': ('1,2of3', '1,2,3of4', '1,2,3,4of5'),
    'Expand': ('1of2',),
}


def _conv_output_shape(xin, pad, ksize, stride, dilation):
    return int((xin + pad - dilation * (ksize - 1) - 1) / stride + 1)


def _convtranspose_output_shape(xin, output_padding, pad, ksize, stride, dilation):
    return stride * (xin - 1) + output_padding + ((ksize - 1) * dilation + 1) - pad


def _auto_pad_valid_shape_calc(x, ksize, stride):
    return math.ceil((x - ksize + 1) / stride)


def _auto_pad_same_shape_calc(x, stride):
    return math.ceil((x) / stride)


def _pooling_shape_calc(inshape, pad, kshape, dilation, stride, ceilmode):
    outshape = (inshape + pad - ((kshape - 1) * dilation + 1)) / stride + 1
    if ceilmode:
        return math.ceil(outshape)
    return math.floor(outshape)


def _axes_neg2pos(len, axes):
    newaxes = []
    for axis in axes:
        if axis < 0:
            newaxes.append(len + axis)
        else:
            newaxes.append(axis)
    return newaxes


def _max_shape(shapes: [numpy.ndarray]):
    maxshape = shapes[0]
    maxvol = volume(maxshape)
    for shape in shapes:
        vol = volume(shape)
        if vol > maxvol:
            maxshape = shape
            maxvol = vol
        elif vol == maxvol:
            if len(shape) > len(maxshape):
                maxshape = shape
    return maxshape


def _contains_shape_tensor(n):
    nodeset = _SHAPE_TENSORS.keys()
    shape_tensors = []
    if n.op_type in nodeset:
        tensor_descs = _SHAPE_TENSORS[n.op_type]
        for desc in tensor_descs:
            strs = desc.split('of')
            indice = strs[0]
            count = int(strs[1])
            if len(n.input) == count:
                indistr = indice.split(',')
                for istr in indistr:
                    shape_tensors.append(n.input[int(istr)])
    return shape_tensors


def _get_shape(item):
    if isinstance(item, numpy.ndarray):
        return list(item.shape)
    elif isinstance(item, (list, tuple)):
        return list(item)
    return [1]


def _get_tensor(item):
    if isinstance(item, numpy.ndarray):
        return item
    elif isinstance(item, (list, tuple)):
        return create_ndarray_f32(item)


class Node():
    def __init__(self, n: onnx.NodeProto):
        self.name = n.name
        self.op_type = n.op_type
        self.nextnodes = []
        self.prevnodes = []
        self.output = []
        self.input = []
        self.proto = n
        self.shape_calc = False
        self.attr = {}
        for att in n.attribute:
            self.attr[att.name] = onnx.helper.get_attribute_value(att)
            self.__setattr__(att.name, get_attribute_data(att))
            if att.name == 'axes':
                if isinstance(self.axes, list):
                    self.axes = tuple(self.axes)

    def set_attr(self, key, val):
        self.attr[key] = val
        self.__setattr__(key, val)

    def add_default_value(self, attname, defaultvalue):
        if not hasattr(self, attname):
            setattr(self, attname, defaultvalue)

    def make_nodeproto(self):
        return onnx.helper.make_node(self.op_type, self.input, self.output, self.name, **self.attr)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        self.value_infer(intensors, outtensors)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        raise NotImplementedError(f'this Node {self.op_type}-{self.name} has no value_infer')

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [0, 0]


class FusedBase(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_tensor(intensors[0].get_numpy())

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [0, 0]


class PWNode(Node):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = ADD_MACS
        self.ratio = max(1, len(self.input) - 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshapes = []
        for tensor in intensors:
            inshapes.append(tensor.get_shape())
        outtensors[0].update_shape(_max_shape(inshapes))
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outshape = outtensors[0].get_shape()
        macs = volume(outshape) * self.ratio * self.op_mac
        return [macs, 0]


class NpMathBase(Node):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = ADD_MACS
        self.ratio = max(1, len(self.input) - 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        maxlen = 0
        for tensor in intensors:
            shape = tensor.get_shape()
            if len(shape) > maxlen:
                maxlen = len(shape)
        inshapes = []
        for tensor in intensors:
            shape = tensor.get_shape()
            for i in range(0, maxlen - len(shape)):
                shape = [1, ] + shape
            inshapes.append(shape)
        outshape = []
        for i in range(maxlen):
            maxdim = 0
            for shape in inshapes:
                if shape[i] > maxdim:
                    maxdim = shape[i]
            outshape.append(maxdim)
        outtensors[0].update_shape(outshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outshape = outtensors[0].get_shape()
        macs = volume(outshape) * self.ratio * self.op_mac
        return [macs, 0]


@NODE_REGISTRY.register()
class SubNode(NpMathBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy() - intensors[1].get_numpy()
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class AddNode(NpMathBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy() + intensors[1].get_numpy()
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class MinNode(NpMathBase):
    def __init__(self, node):
        super().__init__(node)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy()
        for i in range(1, len(intensors)):
            result = not numpy.minimum(result, intensors[i].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class MaxNode(NpMathBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy()
        for i in range(1, len(intensors)):
            result = numpy.maximum(result, intensors[i].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class NegNode(NpMathBase):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_tensor(-intensors[0].get_numpy())


@NODE_REGISTRY.register()
class DivNode(NpMathBase):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = DIV_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if intensors[0].dtype == intensors[1].dtype:
            if intensors[0].dtype in [numpy.int64]:
                result = intensors[0].get_numpy() // intensors[1].get_numpy()
            else:
                result = intensors[0].get_numpy() / intensors[1].get_numpy()
        else:
            result = intensors[0].get_numpy() / intensors[1].get_numpy()
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class MulNode(NpMathBase):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = MUL_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy() * intensors[1].get_numpy()
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class AbsNode(NpMathBase):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.abs(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class CeilNode(NpMathBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.ceil(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class ExpNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = EXP_MACS
        self.ratio = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.exp(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class SoftmaxNode(ExpNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = EXP_MACS + DIV_MACS
        self.ratio = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xexp = numpy.exp(intensors[0].get_numpy())
        result = xexp / numpy.sum(xexp, axis=self.axis, keepdims=True)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class LogNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = LOG_MACS
        self.ratio = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.log(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class ImageScalerNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = ADD_MACS + MUL_MACS
        self.ratio = 1


@NODE_REGISTRY.register()
class InstanceNormalizationNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = ADD_MACS + MUL_MACS + ADD_MACS + DIV_MACS


@NODE_REGISTRY.register()
class SqrtNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = SQRT_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.sqrt(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class PowNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = POW_MACS

    def value_infer(sel, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.power(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class SinNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = SIN_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.sin(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class CosNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = COS_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.cos(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class RangeNode(Node):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        start = intensors[0].get_numpy()
        limit = intensors[1].get_numpy()
        delta = intensors[2].get_numpy()
        result = numpy.arange(start, limit, delta, dtype=intensors[0].dtype)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class SigmoidNode(ExpNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = 1 / (1 + numpy.exp(-intensors[0].get_numpy()))
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class TanhNode(PWNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = TANH_MACS
        self.ratio = 2

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.tanh(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class AtanNode(TanhNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = ATAN_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.arctan(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class SignNode(PWNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.sign(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class HardSigmoidNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = MUL_MACS + ADD_MACS + CMP_MACS * 2
        self.add_default_value('alpha', 0.2)
        self.add_default_value('beta', 0.5)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = max(0, min(1, self.alpha * intensors[0].get_numpy() + self.beta))
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class HardSwishNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = MUL_MACS * 2 + ADD_MACS + CMP_MACS * 2
        self.add_default_value('alpha', 1 / 6)
        self.add_default_value('beta', 0.5)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = intensors[0].get_numpy() * max(0, min(1, self.alpha * intensors[0].get_numpy() + self.beta))
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class ReluNode(PWNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.clip(intensors[0].get_numpy(), 0, None)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class PReluNode(PWNode):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = CMP_MACS + MUL_MACS
        self.ratio = 1


@NODE_REGISTRY.register()
class LeakyReluNode(PWNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = self.op_mac = MUL_MACS + CMP_MACS
        self.add_default_value('alpha', 0.01)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        x = intensors[0].get_numpy()
        outtensor = numpy.zeros_like(x)
        for i in numpy.ndindex(x.shape):
            x = x[i]
            outtensor[i] = x if x >= 0 else x * self.alpha
        outtensors[0].update_tensor(outtensor)


@NODE_REGISTRY.register()
class SumNode(PWNode):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = ADD_MACS
        self.ratio = len(nodeproto.input) - 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        y = intensors[0].get_numpy()
        for i in range(1, len(intensors)):
            y = y + intensors[i].get_numpy()
        outtensors[0].update_tensor(y)


@NODE_REGISTRY.register()
class NonMaxSuppressionNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(intensors) >= 3:
            max_output_boxes_per_class = int(intensors[2].get_numpy()[0])
            outtensors[0].update_shape([max_output_boxes_per_class, 3])
            outtensors[0].update_dtype(numpy.int64)
            return
        raise NotImplementedError()


@NODE_REGISTRY.register()
class LRNNode(PWNode):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 0
        outvol = volume(outtensors[0].get_shape())
        outvol *= (DIV_MACS + EXP_MACS + ADD_MACS + self.size * MUL_MACS)
        macs += outvol
        return [macs, 0]


@NODE_REGISTRY.register()
class LessNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(numpy.bool_)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.less(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [volume(outtensors[0].get_shape()) * CMP_MACS, 0]


@NODE_REGISTRY.register()
class LessOrEqualNode(LessNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.less_equal(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class NotNode(LessNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.logical_not(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class AndNode(LessNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.logical_and(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class WhereNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        cond_shape = intensors[0].get_shape()
        x_shape = intensors[1].get_shape()
        y_shape = intensors[2].get_shape()
        outtensors[0].update_shape(_max_shape((cond_shape, x_shape, y_shape)))
        outtensors[0].update_dtype(intensors[1].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.where(intensors[0].get_numpy(), intensors[1].get_numpy(), intensors[2].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class TransposeNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('perm', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        yshape = []
        if self.perm is None:
            yshape = xshape[::-1]
        else:
            for axis in self.perm:
                yshape.append(xshape[axis])
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.transpose(intensors[0].get_numpy(), self.perm)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class GemmNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('transA', 0)
        self.add_default_value('transB', 0)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        if self.__class__ == GemmNode:
            if self.transA > 0:
                xshape = xshape[::-1]
            else:
                xshape = xshape
            if self.transB > 0:
                yshape = xshape[:-1] + [wshape[-2], ]
            else:
                yshape = xshape[:-1] + [wshape[-1], ]
        else:
            # broadcast support
            batchshape = xshape[:-2] if len(xshape) > len(wshape) else wshape[:-2]
            yshape = batchshape + [xshape[-2], wshape[-1]]
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if self.__class__ == MatMulNode:
            ashape = intensors[0].get_shape()
            bshape = intensors[1].get_shape()
            assert (ashape[-1] == bshape[-2])
            result = numpy.matmul(intensors[0].get_numpy(), intensors[1].get_numpy())
            outtensors[0].update_tensor(result)
        if self.transA > 0:
            A = numpy.transpose(intensors[0].get_numpy())
        else:
            A = intensors[0].get_numpy()
        if self.transB > 0:
            B = numpy.transpose(intensors[1].get_numpy())
        else:
            B = intensors[1].get_numpy()
        C = numpy.matmul(A, B)
        if len(intensors) > 2:
            C = numpy.add(C, intensors[2].get_numpy())
        outtensors[0].update_tensor(C)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        yshape = outtensors[0].get_shape()
        if len(intensors) >= 2:
            weight_shape = intensors[1].get_shape()
            macs = volume(yshape)
            if self.__class__ == GemmNode:
                if self.transB > 0:
                    macs *= weight_shape[-1]
                else:
                    macs *= weight_shape[-2]
            else:
                macs *= weight_shape[-2]
            if len(intensors) == 3:
                macs += volume(yshape) * ADD_MACS
        else:
            raise NotImplementedError()
        return [macs, 0]


@NODE_REGISTRY.register()
class MatMulNode(GemmNode):
    pass


@NODE_REGISTRY.register()
class TileNode(Node):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        input = intensors[0].get_numpy()
        repeats = intensors[1].get_numpy()
        output = numpy.tile(input, repeats)
        outtensors[0].update_tensor(output)


@NODE_REGISTRY.register()
class GatherNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axis', 0)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        idxshape = intensors[1].get_shape()
        axis = _axes_neg2pos(len(xshape), [self.axis])[0]
        yshape = []
        for i in range(len(xshape)):
            if i == axis:
                yshape.extend(idxshape)
            else:
                yshape.append(xshape[i])
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        out = numpy.take(intensors[0].get_numpy(), intensors[1].get_numpy(), axis=self.axis)
        outtensors[0].update_tensor(out)


@NODE_REGISTRY.register()
class ClipNode(PWNode):
    def __init__(self, n):
        super().__init__(n)
        self.op_mac = CMP_MACS * 2
        self.ratio = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        y = numpy.clip(intensors[0].get_numpy(), intensors[1].get_numpy(), intensors[2].get_numpy())
        outtensors[0].update_tensor(y)


@NODE_REGISTRY.register()
class ReciprocalNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = DIV_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.reciprocal(intensors[0].get_numpy())
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class Relu6Node(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = CMP_MACS * 2
        self.ratio = 1

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.clip(intensors[0].get_numpy(), 0, 6)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class ConstantNode(Node):
    def __init__(self, n):
        super().__init__(n)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(self.value.shape)
        outtensors[0].update_dtype(self.value.dtype.type)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_tensor(self.value)


@NODE_REGISTRY.register()
class ConcatNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outshape = intensors[0].get_shape()
        for i in range(len(intensors) - 1):
            shape = intensors[i + 1].get_shape()
            outshape[self.axis] += shape[self.axis]
        outtensors[0].update_shape(outshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        ins = []
        for tensor in intensors:
            ins.append(tensor.get_numpy())
        outtensor = numpy.concatenate(ins, self.axis)
        outtensors[0].update_tensor(outtensor)


# copy from https://github.com/onnx/onnx/blob/main/onnx/backend/test/case/node/onehot.py
def one_hot(indices, depth, axis=-1, dtype=numpy.float32):  # type: ignore
    ''' Compute one hot from indices at a specific axis '''
    values = numpy.asarray(indices)
    rank = len(values.shape)
    depth_range = numpy.arange(depth)
    if axis < 0:
        axis += (rank + 1)
    ls = values.shape[0:axis]
    rs = values.shape[axis:rank]
    targets = numpy.reshape(depth_range, (1,) * len(ls) + depth_range.shape + (1,) * len(rs))
    values = numpy.reshape(numpy.mod(values, depth), ls + (1,) + rs)
    return numpy.asarray(targets == values, dtype=dtype)


@NODE_REGISTRY.register()
class OneHotNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axis', -1)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        indices = intensors[0].get_numpy()
        depth = intensors[1].get_numpy()
        values = intensors[2].get_numpy()
        y = one_hot(indices, depth, self.axis)
        outtensors[0].update_tensor(y)


@NODE_REGISTRY.register()
class TriluNode(FusedBase):
    def __init__(self, n):
        super().__init__(n)
        self.add_default_value('upper', 1)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(intensors) == 2:
            k = intensors[1].get_numpy()
        else:
            k = numpy.array(0).astype(numpy.int64)
        if self.upper == 0:
            result = numpy.tril(intensors[0].get_numpy(), k)
        else:
            result = numpy.triu(intensors[0].get_numpy(), k)
        outtensors[0].update_tensor(result)


@NODE_REGISTRY.register()
class EinsumNode(Node):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        strs = self.equation.split(b',')
        self.ashape = strs[0].replace(b' ', b'')
        strs = strs[1].split(b'->')
        self.bshape = strs[0].replace(b' ', b'')
        self.cshape = strs[1].replace(b' ', b'')

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        shape = []
        map = {}
        shape0 = intensors[0].get_shape()
        shape1 = intensors[1].get_shape()
        for i, v in enumerate(shape0):
            map[self.ashape[i]] = v
        for i, v in enumerate(shape1):
            map[self.bshape[i]] = v
        for k in self.cshape:
            shape.append(map[k])
        outtensors[0].update_shape(shape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 1
        map = {}
        shape0 = intensors[0].get_shape()
        shape1 = intensors[1].get_shape()
        for i, v in enumerate(shape0):
            map[self.ashape[i]] = v
        for i, v in enumerate(shape1):
            map[self.bshape[i]] = v
        for key in map.keys():
            macs *= map[key]
        return [macs, 0]


@NODE_REGISTRY.register()
class UnsqueezeNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axes', [0])

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        newaxis_len = len(inshape) + len(axes)
        axes = _axes_neg2pos(newaxis_len, axes)
        newshape = []
        idx = 0
        for i in range(newaxis_len):
            if i in axes:
                newshape.append(1)
            else:
                newshape.append(inshape[idx])
                idx += 1
        outtensors[0].update_shape(newshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensor = intensors[0].get_numpy()
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        for axis in axes:
            outtensor = numpy.expand_dims(outtensor, axis=axis)
        outtensors[0].update_tensor(outtensor)


@NODE_REGISTRY.register()
class SqueezeNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axes', [0])

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        outshape = []
        if len(intensors) == 2:
            self.axes = intensors[1].get_numpy()
        axes = _axes_neg2pos(len(inshape), self.axes)
        for i in range(len(inshape)):
            if i in axes:
                continue
            else:
                outshape.append(inshape[i])
        outtensors[0].update_shape(outshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensor = intensors[0].get_numpy().copy()
        idx = 0
        if len(intensors) == 2:
            self.axes = intensors[1].get_numpy()
        for axis in self.axes:
            outtensor = numpy.squeeze(outtensor, axis=axis - idx)
            idx += 1
        outtensors[0].update_tensor(outtensor)


@NODE_REGISTRY.register()
class ShapeNode(Node):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        ret = numpy.array(intensors[0].get_shape(), dtype=numpy.int64)
        outtensors[0].update_tensor(ret)


@NODE_REGISTRY.register()
class ResizeNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        roi = []
        sizes = []
        if len(intensors) == 2:
            scales = intensors[1].get_numpy()
        elif len(intensors) >= 3:
            roi = intensors[1].get_numpy()
            scales = intensors[2].get_numpy()
            if len(intensors) >= 4:
                sizes = intensors[3].get_numpy()

        newshape = []
        if is_valid_ndarray(sizes):
            if len(sizes) == len(xshape):
                newshape = sizes
            if len(sizes) == 2:
                newshape = xshape[:2] + sizes
        else:
            if is_valid_ndarray(scales):
                newshape = []
                for src, scale in zip(xshape, scales):
                    newshape.append(math.floor(src * scale))

        if is_valid_ndarray(newshape):
            if newshape.dtype != numpy.int64:
                newshape = newshape.astype(dtype=numpy.int64)
        outtensors[0].update_shape(list(newshape))
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 0
        outvol = volume(outtensors[0].get_shape())
        if self.mode == b'nearest':
            outvol *= 0
        elif self.mode == b'linear':
            outvol *= RESIZE_LINEAR_MACS
        elif self.mode == b'cubic':
            outvol *= RESIZE_CUBIC_MACS
        macs += outvol
        return [macs, 0]


@NODE_REGISTRY.register()
class UpsampleNode(ResizeNode):
    pass


@NODE_REGISTRY.register()
class PoolBase(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = CMP_MACS
        self.add_default_value('kernel_shape', (3, 3))
        self.add_default_value('ceil_mode', 0)
        self.add_default_value('pads', (0, 0, 0, 0))
        self.add_default_value('strides', (1, 1))
        self.add_default_value('dilations', (1, 1))
        self.add_default_value('auto_pad', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        if self.auto_pad is not None and self.auto_pad != b'NOTSET':
            outshape = inshape[:2]
            if self.auto_pad in [b'SAME_LOWER', b'SAME_UPPER']:
                outshape += (_auto_pad_same_shape_calc(inshape[2], self.strides[0]),)
                if len(self.strides) == 2:
                    outshape += [_auto_pad_same_shape_calc(inshape[3], self.strides[1]), ]
            elif self.auto_pad == b'VALID':
                outshape += [_auto_pad_valid_shape_calc(inshape[2], self.kernel_shape[0], self.strides[0]), ]
                if len(self.strides) == 2:
                    outshape += [_auto_pad_valid_shape_calc(inshape[3], self.kernel_shape[1], self.strides[1]), ]
        else:
            if len(self.kernel_shape) == 1:
                outshape = inshape[:2] + [
                    _pooling_shape_calc(inshape[2], self.pads[0] + self.pads[1], self.kernel_shape[0],
                                        self.dilations[0],
                                        self.strides[0], self.ceil_mode), ]
            if len(self.kernel_shape) == 2:
                outshape = inshape[:2] + [
                    _pooling_shape_calc(inshape[2], self.pads[0] + self.pads[2], self.kernel_shape[0],
                                        self.dilations[0],
                                        self.strides[0],
                                        self.ceil_mode),
                    _pooling_shape_calc(inshape[3], self.pads[1] + self.pads[3], self.kernel_shape[1],
                                        self.dilations[1],
                                        self.strides[1],
                                        self.ceil_mode),
                ]
            if len(self.kernel_shape) == 3:
                outshape = inshape[:2] + [
                    _pooling_shape_calc(inshape[2], self.pads[0] + self.pads[0], self.kernel_shape[0],
                                        self.dilations[0],
                                        self.strides[0],
                                        self.ceil_mode),
                    _pooling_shape_calc(inshape[3], self.pads[1] + self.pads[1], self.kernel_shape[1],
                                        self.dilations[1],
                                        self.strides[1],
                                        self.ceil_mode),
                    _pooling_shape_calc(inshape[4], self.pads[2] + self.pads[2], self.kernel_shape[2],
                                        self.dilations[1],
                                        self.strides[2],
                                        self.ceil_mode),
                ]
        outtensors[0].update_shape(outshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outshape = outtensors[0].get_shape()
        outvol = volume(outshape)
        macs = outvol * CMP_MACS * self.kernel_shape[0]
        if len(self.kernel_shape) == 2:
            macs *= self.kernel_shape[1]
        return [macs, 0]


@NODE_REGISTRY.register()
class AveragePoolNode(PoolBase):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = ADD_MACS


@NODE_REGISTRY.register()
class MaxPoolNode(PoolBase):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = CMP_MACS

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        self.shape_infer(intensors, outtensors)
        ot = outtensors[0].get_numpy()
        for i in numpy.ndindex(ot.shape):
            batch = i[0]
            ocn = i[1]
            oh = i[2]
            ow = i[3]
            t = ot[i]
            ks = tuple(self.kernel_shape)
            for j in numpy.ndindex(ks):
                kh = j[0]
                kw = j[1]
                srch = oh * self.strides[0] + kh - self.pads[0]
                srcw = ow * self.strides[1] + kw - self.pads[1]
                if srch < 0 or srch >= xshape[2] or srcw < 0 or srcw >= xshape[3]:
                    continue
                else:
                    srcv = intensors[0].get_numpy()[batch, ocn, srch, srcw]
                t = max(srcv, t)
            ot[i] = t


@NODE_REGISTRY.register()
class DropoutNode(FusedBase):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)
        if len(outtensors) == 2:
            outtensors[1].update_shape(intensors[0].get_shape())
            outtensors[1].update_dtype(numpy.bool_)


@NODE_REGISTRY.register()
class GlobalAveragePoolNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        shape = inshape[0:2]
        for i in range(2, len(inshape)):
            shape += (1,)
        outtensors[0].update_shape(shape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        x = intensors[0].get_numpy()
        h = x.shape[2]
        w = x.shape[3]
        y = numpy.zeros(x.shape[:2], dtype=numpy.float32)
        for i in numpy.ndindex(y.shape):
            t = 0
            for j in numpy.ndindex((h, w)):
                xi = i + j
                t += x[xi]
            t /= (h * w)
            y[i] = t
        outtensors[0].update_tensor(y)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        outshape = outtensors[0].get_shape()
        macs = volume(inshape) * ADD_MACS + volume(outshape) * DIV_MACS
        return [macs, 0]


@NODE_REGISTRY.register()
class ExpandNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        expandshape = intensors[1].get_numpy().tolist()
        if not isinstance(expandshape, list):
            expandshape = [expandshape, ]
        yshape = []
        if len(xshape) < len(expandshape):
            for i in range(len(xshape), len(expandshape)):
                xshape = [1, ] + xshape
        for x, e in zip(xshape, expandshape):
            yshape.append(max(x, e))
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        output = intensors[0].get_numpy() * numpy.ones(intensors[1].get_numpy(), dtype=intensors[0].dtype)
        outtensors[0].update_tensor(output)


@NODE_REGISTRY.register()
class PadNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('pads', None)
        self.add_default_value('value', 0)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        newshape = []
        if self.pads is None:
            if len(intensors) > 1:
                pads = intensors[1].get_numpy()
                for i, v in enumerate(inshape):
                    newshape.append(v + pads[i] + pads[i + len(inshape)])
        else:
            for i, v in enumerate(inshape):
                newshape.append(v + self.pads[i] + self.pads[i + len(inshape)])
        newshape = [int(val) for val in newshape]
        outtensors[0].update_shape(newshape)
        outtensors[0].update_dtype(intensors[0].dtype)


@NODE_REGISTRY.register()
class IdentityNode(FusedBase):
    pass


@NODE_REGISTRY.register()
class ErfNode(FusedBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensor = numpy.zeros_like(intensors[0].get_numpy())
        for i in numpy.ndindex(intensors[0].shape):
            outtensor[i] = math.erf(intensors[0].get_numpy()[i])
        outtensors[0].update_tensor(outtensor)


@NODE_REGISTRY.register()
class BatchNormalizationNode(FusedBase):
    def __init__(self, n):
        super().__init__(n)
        self.add_default_value('epsilon', 1e-05)
        self.add_default_value('momentum', 0.9)
        self.add_default_value('training_mode', int(0))

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        x = intensors[0].get_numpy()
        scale = intensors[1].get_numpy()
        b = intensors[2].get_numpy()
        mean = intensors[3].get_numpy()
        var = intensors[4].get_numpy()
        y = numpy.zeros_like(x)
        for i in numpy.ndindex(y.shape):
            cn = i[1]
            sqrt_var = math.sqrt(var[cn] + self.epsilon)
            sm = scale[cn] / sqrt_var
            sv = b[cn]
            m = mean[cn]
            y[i] = (x[i] - m) * sm + sv
        outtensors[0].update_tensor(y)

    # Fusion of batchnorm is determined by inference engine, here just gives the MACs.
    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        base = volume(outtensors[0].get_shape())
        base *= ADD_MACS + SQRT_MACS + DIV_MACS + ADD_MACS + MUL_MACS
        return [base, 0]


@NODE_REGISTRY.register()
class FlattenNode(Node):
    def __init__(self, node):
        super().__init__(node)
        self.add_default_value('axis', None)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        x = intensors[0].get_numpy()
        if self.axis is None:
            y = x.reshape((x.shape[0], -1))
        else:
            vol = 1
            for i in range(self.axis):
                vol *= x.shape[i]
            y = x.reshape((vol, -1))
        outtensors[0].update_tensor(y)


# copy from https://github.com/onnx/onnx/blob/main/onnx/backend/test/case/node/argmax.py
def argmax_use_numpy(data: numpy.ndarray, axis: int = 0, keepdims: int = 1) -> (numpy.ndarray):
    result = numpy.argmax(data, axis=axis)
    if (keepdims == 1):
        result = numpy.expand_dims(result, axis)
    return result.astype(numpy.int64)


@NODE_REGISTRY.register()
class ArgMaxNode(Node):
    def __init__(self, n):
        super().__init__(n)
        self.add_default_value('axis', 0)
        self.add_default_value('keepdims', 1)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        out = argmax_use_numpy(data, self.axis, self.keepdims)
        outtensors[0].update_tensor(out)


@NODE_REGISTRY.register()
class ArrayFeatureExtractorNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[1].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)


@NODE_REGISTRY.register()
class ZipMapNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape([intensors[0].get_shape()[0], ])
        outtensors[0].update_dtype(intensors[0].dtype)


@NODE_REGISTRY.register()
class SliceNode(Node):
    def __init__(self, n):
        super(SliceNode, self).__init__(n)
        self.add_default_value('steps', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        inshape = intensors[0].get_shape()
        if len(intensors) == 1:
            starts = self.starts
            ends = self.ends
            axes = self.axes
            if self.steps is None:
                steps = [1] * len(starts)
            else:
                steps = self.steps
        else:
            elesize = len(intensors[1].get_numpy())
            starts = intensors[1].get_numpy()
            ends = intensors[2].get_numpy()
            if len(intensors) == 3:
                # undef beheviour of bidaf-9.onnx
                axes = [0]
                steps = [1]
            else:
                axes = [0] * elesize
                steps = [1] * elesize
            if len(intensors) >= 4:
                axes = intensors[3].get_numpy()
            if len(intensors) >= 5:
                steps = intensors[4].get_numpy()

        axes = _axes_neg2pos(len(inshape), axes)
        newshape = inshape.copy()
        for a in axes:
            newshape[a] = 0
        for s, e, a, st in zip(starts, ends, axes, steps):
            if s < 0:
                s = max(0, inshape[a] + s)
            else:
                s = max(s, 0)
            if e < 0:
                e = max(0, inshape[a] + e)
            else:
                e = min(e, inshape[a])
            tmp = abs(e - s)
            newshape[a] += abs(math.ceil(tmp / st))
        outtensors[0].update_shape(newshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        datashape = intensors[0].get_shape()
        if len(intensors) == 3:
            starts = intensors[1].get_numpy()
            ends = intensors[2].get_numpy()
            x = data[starts[0]:ends[0]]
        if len(intensors) == 4:
            starts = intensors[1].get_numpy()
            ends = intensors[2].get_numpy()
            axes = intensors[3].get_numpy()
            index = 0
            x = data.copy()
            for i in axes:
                if i == 0:
                    x = x[starts[index]:ends[index], ...]
                if i == 1:
                    x = x[:, starts[index]:ends[index], ...]
                if i == 2:
                    x = x[:, :, starts[index]:ends[index], ...]
                if i == 3:
                    x = x[:, :, :, starts[index]:ends[index], ...]
                if i == 4:
                    x = x[:, :, :, :, starts[index]:ends[index], ...]
                index += 1
        if len(intensors) == 5:
            starts = intensors[1].get_numpy()
            ends = intensors[2].get_numpy()
            axes = intensors[3].get_numpy()
            steps = intensors[4].get_numpy()
            index = 0
            x = data.copy()
            for i in axes:
                if i == 0:
                    x = x[starts[index]:ends[index]:steps[index], ...]
                if i == 1:
                    x = x[:, starts[index]:ends[index]:steps[index], ...]
                if i == 2:
                    x = x[:, :, starts[index]:ends[index]:steps[index], ...]
                if i == 3:
                    x = x[:, :, :, starts[index]:ends[index]:steps[index], ...]
                if i == 4:
                    x = x[:, :, :, :, starts[index]:ends[index]:steps[index], ...]
                index += 1
        if len(intensors) == 1:
            index = 0
            x = data.copy()
            for i in self.axes:
                if i == 0:
                    x = x[self.starts[index]:self.ends[index], ...]
                if i == 1:
                    x = x[:, self.starts[index]:self.ends[index], ...]
                if i == 2:
                    x = x[:, :, self.starts[index]:self.ends[index], ...]
                if i == 3:
                    x = x[:, :, :, self.starts[index]:self.ends[index], ...]
                if i == 4:
                    x = x[:, :, :, :, self.starts[index]:self.ends[index], ...]
                index += 1
        outtensors[0].update_tensor(x)


@NODE_REGISTRY.register()
class ReduceMeanNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axes', None)
        self.add_default_value('keepdims', 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        yshape = []
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        if axes is None:
            return [[1]]
        else:
            axes = _axes_neg2pos(len(xshape), axes)

        for i in range(len(xshape)):
            if i in axes:
                if self.keepdims:
                    yshape.append(1)
            else:
                yshape.append(xshape[i])
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        reduced = numpy.mean(intensors[0].get_numpy(), axis=axes, keepdims=self.keepdims == 1)
        outtensors[0].update_tensor(reduced)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        vol = volume(intensors[0].get_shape())
        return [vol * ADD_MACS, 0]


@NODE_REGISTRY.register()
class ReduceProdNode(ReduceMeanNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        reduced = numpy.prod(intensors[0].get_numpy(), axis=axes, keepdims=self.keepdims == 1)
        outtensors[0].update_tensor(reduced)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        datashape = intensors[0].get_shape()
        vol = volume(datashape)
        return [vol * MUL_MACS, 0]


@NODE_REGISTRY.register()
class ReduceSumNode(ReduceMeanNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(intensors) == 2:
            axes = tuple(intensors[1].get_numpy().tolist())
        else:
            axes = self.axes
        reduced = numpy.sum(intensors[0].get_numpy(), axis=axes, keepdims=self.keepdims == 1)
        outtensors[0].update_tensor(reduced)


@NODE_REGISTRY.register()
class ReduceMinNode(ReduceMeanNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        reduced = numpy.minimum.reduce(data, axis=axes, keepdims=self.keepdims == 1)
        outtensors[0].update_tensor(reduced)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        datashape = intensors[0].get_shape()
        vol = volume(datashape)
        return [vol * CMP_MACS, 0]


@NODE_REGISTRY.register()
class ReduceMaxNode(ReduceMinNode):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        if len(intensors) == 2:
            axes = intensors[1].get_numpy()
        else:
            axes = self.axes
        reduced = numpy.maximum.reduce(data, axis=axes, keepdims=self.keepdims == 1)
        outtensors[0].update_tensor(reduced)


@NODE_REGISTRY.register()
class TopKNode(Node):
    def __init__(self, node):
        super().__init__(node)
        self.add_default_value('axis', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        k = intensors[1].get_numpy()[0]
        # when the input tensor only contain 1 dimension, the axis attribute (default: 0) may not appear in the node
        if len(xshape) == 1 and self.axis is None:
            self.axis = 0
        newshape = []
        for i in range(len(xshape)):
            if i == self.axis:
                newshape.append(k)
            else:
                newshape.append(xshape[i])
        outtensors[0].update_shape(newshape)
        outtensors[0].update_dtype(intensors[0].dtype)
        if len(outtensors) == 2:
            outtensors[1].update_shape(newshape)
            outtensors[1].update_dtype(numpy.int64)


@NODE_REGISTRY.register()
class ScanNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('num_scan_inputs', None)
        self.add_default_value('scan_input_directions', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if len(self.output) == 2:
            outtensors[0].update_shape(intensors[3].get_shape())
            outtensors[0].update_dtype(intensors[3].dtype)
            outtensors[1].update_shape(intensors[3].get_shape())
            outtensors[1].update_dtype(intensors[3].dtype)
        # TODO
        outtensors[0].update_shape([1, 1])
        outtensors[0].update_dtype(intensors[3].dtype)
        outtensors[1].update_shape([1, 1])
        outtensors[1].update_dtype(intensors[3].dtype)
        outtensors[2].update_shape([1, ])
        outtensors[2].update_dtype(intensors[3].dtype)
        outtensors[3].update_shape(intensors[3].get_shape())
        outtensors[3].update_dtype(intensors[3].dtype)
        outtensors[4].update_shape(intensors[3].get_shape())
        outtensors[4].update_dtype(intensors[3].dtype)


@NODE_REGISTRY.register()
class CompressNode(Node):
    def __init__(self, node):
        super().__init__(node)
        self.add_default_value('axis', None)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        ret = numpy.compress(intensors[1].get_numpy(), intensors[0].get_numpy(), self.axis)
        outtensors[0].update_tensor(ret)


@NODE_REGISTRY.register()
class HardmaxNode(PWNode):
    pass


@NODE_REGISTRY.register()
class CategoryMapperNode(PWNode):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.op_mac = 0


@NODE_REGISTRY.register()
class LSTMNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('direction', None)
        self.add_default_value('hidden_size', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        seq_len = xshape[0]
        batch = xshape[1]
        num_dir = wshape[0]
        h_len = wshape[1] // 4
        outtensors[0].update_shape([seq_len, num_dir, batch, h_len])
        outtensors[0].update_dtype(intensors[0].dtype)
        if len(outtensors) > 1:
            outtensors[1].update_shape([num_dir, batch, h_len])
            outtensors[1].update_dtype(intensors[0].dtype)
            if len(outtensors) > 2:
                outtensors[2].update_shape([num_dir, batch, h_len])
                outtensors[2].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        wshape = intensors[1].get_shape()
        rshape = intensors[2].get_shape()
        bshape = intensors[3].get_shape()
        batch = intensors[0].get_shape()[1]
        macs = volume(wshape) + volume(rshape) + volume(bshape) * ADD_MACS
        macs *= batch
        return [macs, 0]


@NODE_REGISTRY.register()
class ConvNode(Node):
    def __init__(self, n):
        super(ConvNode, self).__init__(n)
        self.add_default_value('auto_pad', None)
        self.add_default_value('pads', (0, 0, 0, 0))
        self.add_default_value('strides', (1, 1))
        self.add_default_value('dilations', (1, 1))
        self.add_default_value('group', 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        shape = []
        if self.auto_pad is not None and self.auto_pad != b'NOTSET':
            if self.auto_pad in [b'SAME_LOWER', b'SAME_UPPER']:
                shape = (xshape[0], wshape[0], math.ceil(xshape[2] / self.strides[0]))
                if len(xshape) == 4:
                    shape += (math.ceil(xshape[3] / self.strides[1]),)
            elif self.auto_pad == b'VALID':
                oh = math.ceil((xshape[2] - wshape[2] + 1) / self.strides[0])
                shape = (xshape[0], wshape[0], oh)
                if len(xshape) == 4:
                    ow = math.ceil((xshape[3] - wshape[3] + 1) / self.strides[1])
                    shape += (ow,)
            else:
                assert 0

        else:
            if len(xshape) == 5:
                od = _conv_output_shape(xshape[2], self.pads[0] + self.pads[3], wshape[2], self.strides[0],
                                        self.dilations[0])
                oh = _conv_output_shape(xshape[3], self.pads[1] + self.pads[4], wshape[3], self.strides[1],
                                        self.dilations[1])
                ow = _conv_output_shape(xshape[4], self.pads[2] + self.pads[5], wshape[4], self.strides[2],
                                        self.dilations[2])
                shape = (xshape[0], wshape[0], od, oh, ow)
            if len(xshape) == 4:
                oh = _conv_output_shape(xshape[2], self.pads[0] + self.pads[2], wshape[2], self.strides[0],
                                        self.dilations[0])
                ow = _conv_output_shape(xshape[3], self.pads[1] + self.pads[3], wshape[3], self.strides[1],
                                        self.dilations[1])
                shape = (xshape[0], wshape[0], oh, ow)
            elif len(xshape) == 3:
                oh = _conv_output_shape(xshape[2], self.pads[0] + self.pads[1], wshape[2], self.strides[0],
                                        self.dilations[0])
                shape = (xshape[0], wshape[0], oh)
        outtensors[0].update_shape(shape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if self.group != 1:
            raise NotImplementedError()
        self.shape_infer(intensors, outtensors)
        outshape = outtensors[0].get_shape()
        outtensor = outtensors[0].get_numpy()
        has_bias = len(intensors) > 2
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        if len(wshape) != 4:
            raise NotImplementedError()

        reduce_shape = tuple(wshape[1:])
        for i in numpy.ndindex(tuple(outshape)):
            batch = i[0]
            ocn = i[1]
            oh = i[2]
            ow = i[3]
            t = outtensor[i]
            if has_bias:
                t = intensors[2].get_numpy()[ocn]
            for j in numpy.ndindex(reduce_shape):
                icn = j[0]
                kh = j[1]
                kw = j[2]
                srch = oh * self.strides[0] + kh * self.dilations[0] - self.pads[0]
                srcw = ow * self.strides[1] + kw * self.dilations[1] - self.pads[1]
                if srch < 0 or srch >= xshape[2] or srcw < 0 or srcw >= xshape[3]:
                    srcv = 0
                else:
                    srcv = intensors[0].get_numpy()[batch, icn, srch, srcw]
                wv = intensors[1].get_numpy()[(ocn,) + j]
                t += srcv * wv
            outtensor[i] = t
        outtensors[0].update_tensor(outtensor)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 0
        if len(outtensors) == 1:
            if len(intensors) == 3 or len(intensors) == 2:
                kernel_shape = intensors[1].get_shape()
                outvol = volume(outtensors[0].get_shape())
                reduce_shape = kernel_shape[1:]
                reduce_vol = volume(reduce_shape)
                macs += outvol * reduce_vol * MUL_MACS
                if len(intensors) > 2:
                    macs += (outvol * ADD_MACS)
        return [macs, 0]


@NODE_REGISTRY.register()
class ReduceL2Node(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axes', None)
        self.add_default_value('keepdims', 1)
        self.axes = tuple(self.axes) if self.axes is not None else None

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        reduced = numpy.sqrt(
            numpy.sum(intensors[0].get_numpy() * intensors[0].get_numpy(), axis=self.axes, keepdims=self.keepdims == 1))
        outtensors[0].update_tensor(reduced)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        vol = volume(intensors[0].get_shape())
        return [vol * (ADD_MACS + SQRT_MACS), 0]


@NODE_REGISTRY.register()
class CumSumNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = ADD_MACS
        self.ratio = 1
        self.add_default_value('exclusive', 0)
        self.add_default_value('reverse', 0)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        if self.exclusive == 0 and self.reverse == 0:
            y = numpy.cumsum(intensors[0].get_numpy(), intensors[1].get_numpy())
            outtensors[0].update_tensor(y)
            return
        raise NotImplementedError(f"CumSum doesnt support {self.exclusive} {self.reverse}")


@NODE_REGISTRY.register()
class NonZeroNode(Node):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        condi = intensors[0].get_numpy()
        result = numpy.array(numpy.nonzero(condi), dtype=numpy.int64)
        if volume(result.shape) == 0:
            condi = numpy.ones_like(intensors[0].get_numpy())
            result = numpy.array(numpy.nonzero(condi), dtype=numpy.int64)
        outtensors[0].update_tensor(result)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [volume(outtensors[0].get_shape()) * CMP_MACS, 0]


@NODE_REGISTRY.register()
class EqualNode(Node):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.equal(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [volume(outtensors[0].get_shape()) * CMP_MACS, 0]


@NODE_REGISTRY.register()
class FloorNode(FusedBase):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        ret = numpy.floor(intensors[0].get_numpy())
        outtensors[0].update_tensor(ret)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        return [volume(outtensors[0].get_shape()) * CMP_MACS, 0]


@NODE_REGISTRY.register()
class RoiAlignNode(Node):
    def __init__(self, node):
        super().__init__(node)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        if len(xshape) == 4 and self.output_height is not None and self.output_width is not None:
            newshape = xshape[:2] + [self.output_height, self.output_width]
            outtensors[0].update_shape(newshape)
            outtensors[0].update_dtype(intensors[0].dtype)
        else:
            raise NotImplementedError()


@NODE_REGISTRY.register()
class ScatterElementsNode(Node):
    def __init__(self, node):
        super().__init__(node)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)


# copy from https://github.com/onnx/onnx/blob/main/onnx/backend/test/case/node/scatternd.py
def scatter_nd_impl(data, indices, updates, reduction='none'):  # type: ignore

    # Check tensor shapes
    assert indices.shape[-1] <= len(data.shape)
    assert updates.shape == indices.shape[:-1] + data.shape[indices.shape[-1]:]

    # Compute output
    output = numpy.copy(data)
    for i in numpy.ndindex(indices.shape[:-1]):
        rawoutput = output[tuple(indices[i])]
        # NOTE: The order of iteration in this loop is not specified.
        if reduction == "add":
            rawoutput += updates[i]
        elif reduction == "mul":
            rawoutput *= updates[i]
        elif reduction == "max":
            rawoutput = numpy.maximum(rawoutput, updates[i])
        elif reduction == "min":
            rawoutput = numpy.minimum(rawoutput, updates[i])
        else:
            rawoutput = updates[i]
        output[tuple(indices[i])] = rawoutput
    return output


# copy from https://github.com/onnx/onnx/blob/main/onnx/backend/test/case/node/gathernd.py
def gather_nd_impl(
        data: numpy.ndarray, indices: numpy.ndarray, batch_dims: int
) -> numpy.ndarray:
    # Note the data rank - will be reused multiple times later
    data_rank = len(data.shape)

    # Check input tensors' shape/rank condition
    assert indices.shape[-1] <= data_rank

    # The list of data/indice shape of batch_dims
    batch_dims_shape = []

    # The number of elements in the batch_dims for data/indice array
    batch_dims_size = 1

    # Check the shape of indice and data are identicial for batch dims.
    for i in range(batch_dims):
        batch_dims_shape.append(indices.shape[i])
        batch_dims_size *= indices.shape[i]

    # Compute output of the op as below

    # Compute shape of output array
    output_shape = (
        batch_dims_shape + list(indices.shape)[batch_dims:-1]
        if (indices.shape[-1] == data_rank - batch_dims)
        else batch_dims_shape
             + list(indices.shape)[batch_dims:-1]
             + list(data.shape)[batch_dims + indices.shape[-1]:]
    )

    # Placeholder for output data
    output_data_buffer = []

    # Flatten 'indices' to 2D array
    reshaped_indices = indices.reshape(batch_dims_size, -1, indices.shape[-1])

    # Flatten 'data' to array of shape (batch_dim_size, data.shape[batch_dimes:])
    reshaped_data = data.reshape((batch_dims_size,) + data.shape[batch_dims:])

    # gather each scalar value from 'data'
    for batch_dim in range(reshaped_indices.shape[0]):
        for outer_dim in range(reshaped_indices.shape[1]):
            gather_index = tuple(reshaped_indices[batch_dim][outer_dim])
            output_data_buffer.append(reshaped_data[(batch_dim, *gather_index)])
    return numpy.asarray(output_data_buffer, dtype=data.dtype).reshape(output_shape)


@NODE_REGISTRY.register()
class ScatterNDNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        indices = intensors[1].get_numpy()
        updates = intensors[2].get_numpy()
        outtensors[0].update_tensor(scatter_nd_impl(data, indices, updates))


@NODE_REGISTRY.register()
class GatherNDNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data_shape = intensors[0].get_shape()
        indice_shape = intensors[1].get_shape()
        batch_dims = 0
        # Note the data rank - will be reused multiple times later
        data_rank = len(data_shape)

        # Check input tensors' shape/rank condition
        assert indice_shape[-1] <= data_rank

        # The list of data/indice shape of batch_dims
        batch_dims_shape = []

        # The number of elements in the batch_dims for data/indice array
        batch_dims_size = 1

        # Check the shape of indice and data are identicial for batch dims.
        for i in range(batch_dims):
            batch_dims_shape.append(indice_shape[i])
            batch_dims_size *= indice_shape[i]

        # Compute output of the op as below

        # Compute shape of output array
        output_shape = (
            batch_dims_shape + list(indice_shape)[batch_dims:-1]
            if (indice_shape[-1] == data_rank - batch_dims)
            else batch_dims_shape
                 + list(indice_shape)[batch_dims:-1]
                 + list(data_shape)[batch_dims + indice_shape[-1]:]
        )
        outtensors[0].update_shape(output_shape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        data = intensors[0].get_numpy()
        indices = intensors[1].get_numpy()
        ret = gather_nd_impl(data, indices, 0)
        outtensors[0].update_tensor(ret)


@NODE_REGISTRY.register()
class RandomUniformLikeNode(Node):
    def __init__(self, n):
        super().__init__(n)
        self.add_default_value('dtype', None)
        self.add_default_value('high', 1.0)
        self.add_default_value('low', 0.0)
        self.add_default_value('seed', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        if self.dtype is None:
            outtensors[0].update_dtype(intensors[0].dtype)
        else:
            outtensors[0].update_dtype(onnxdtype2npdtype(self.dtype))


@NODE_REGISTRY.register()
class RandomNormalLikeNode(RandomUniformLikeNode):
    pass


@NODE_REGISTRY.register()
class GreaterNode(Node):
    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        result = numpy.greater(intensors[0].get_numpy(), intensors[1].get_numpy())
        outtensors[0].update_tensor(result)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outshape = outtensors[0].get_shape()
        return [volume(outshape) * CMP_MACS, 0]


@NODE_REGISTRY.register()
class DequantizeLinearNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = MUL_MACS + ADD_MACS
        self.ratio = 1

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[1].dtype)


@NODE_REGISTRY.register()
class LayerNormalizationNode(Node):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.add_default_value('axis', -1)
        self.add_default_value('epsilon ', 1e-05)
        self.add_default_value('stash_type', 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        tshape = intensors[0].get_shape()
        axis = _axes_neg2pos(len(tshape), [self.axis])[0]
        vol = volume(tshape)
        tshape[axis] = 1
        vol2 = volume(tshape)
        return [vol * (MUL_MACS * 3 + +ADD_MACS * 4) + vol2 * (ADD_MACS + SQRT_MACS + DIV_MACS), 0]


@NODE_REGISTRY.register()
class QuantizeLinearNode(PWNode):
    def __init__(self, node_proto):
        super().__init__(node_proto)
        self.op_mac = MUL_MACS + ADD_MACS
        self.ratio = 1

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[2].dtype)


@NODE_REGISTRY.register()
class MatMulIntegerNode(GemmNode):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        super().shape_infer(intensors, outtensors)
        outtensors[0].update_dtype(numpy.int32)


@NODE_REGISTRY.register()
class QLinearMatMulNode(GemmNode):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('transA', None)
        self.add_default_value('transB', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[3].get_shape()

        if self.__class__ == GemmNode:
            if self.transA is not None and self.transA > 0:
                xshape = xshape[::-1]
            else:
                xshape = xshape
            if self.transB is not None and self.transB > 0:
                yshape = xshape[:-1] + [wshape[-2], ]
            else:
                yshape = xshape[:-1] + [wshape[-1], ]
        else:
            yshape = xshape[:-1] + [wshape[-1], ]
        outtensors[0].update_shape(yshape)
        outtensors[0].update_dtype(intensors[-1].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        weight_shape = intensors[3].get_shape()
        macs = volume(xshape)
        if self.__class__ == GemmNode:
            macs *= weight_shape[0]
        else:
            macs *= weight_shape[-1]
        return [macs, 0]


@NODE_REGISTRY.register()
class QLinearConvNode(ConvNode):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[3].get_shape()
        shape = []
        if self.auto_pad is not None and self.auto_pad != b'NOTSET':
            if self.auto_pad in [b'SAME_LOWER', b'SAME_UPPER']:
                shape = [xshape[0], wshape[0], math.ceil(xshape[2] / self.strides[0])]
                if len(xshape) == 4:
                    shape += [math.ceil(xshape[3] / self.strides[1]), ]
        else:
            if len(xshape) == 4:
                oh = _conv_output_shape(xshape[2], self.pads[0] + self.pads[2], wshape[2], self.strides[0],
                                        self.dilations[0])
                ow = _conv_output_shape(xshape[3], self.pads[1] + self.pads[3], wshape[3], self.strides[1],
                                        self.dilations[1])
                shape = [xshape[0], wshape[0], oh, ow]
            elif len(xshape) == 3:
                oh = _conv_output_shape(xshape[2], self.pads[0] + self.pads[1], wshape[2], self.strides[0],
                                        self.dilations[0])
                shape = [xshape[0], wshape[0], oh]
        outtensors[0].update_shape(shape)
        outtensors[0].update_dtype(intensors[-2].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 0
        outshape = outtensors[0].get_shape()
        if len(outtensors) == 1:
            kernel_shape = intensors[3].get_shape()
            if len(kernel_shape) > 3:
                outvol = volume(outshape)
                macs += outvol * kernel_shape[1] * kernel_shape[2] * kernel_shape[3]
            elif len(kernel_shape) == 3:
                outvol = volume(outshape)
                macs += outvol * kernel_shape[1] * kernel_shape[2]
            else:
                outvol = 0
                raise NotImplementedError()
            if len(intensors) == 9:
                macs += (outvol * ADD_MACS)
        return [macs, 0]


@NODE_REGISTRY.register()
class ConvTransposeNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('pads', (0, 0, 0, 0))
        self.add_default_value('output_padding', (0, 0, 0, 0))
        self.add_default_value('strides', (1, 1))
        self.add_default_value('dilations', (1, 1))
        self.add_default_value('output_shape', (0, 0))
        self.add_default_value('group', 1)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        shape = []
        outc = self.group * wshape[1]
        if len(xshape) == 5:
            od = _convtranspose_output_shape(xshape[2], self.output_padding[0], self.pads[0] + self.pads[3], wshape[2],
                                             self.strides[0],
                                             self.dilations[0])
            ow = _convtranspose_output_shape(xshape[3], self.output_padding[1], self.pads[1] + self.pads[4], wshape[3],
                                             self.strides[1],
                                             self.dilations[1])
            oh = _convtranspose_output_shape(xshape[4], self.output_padding[2], self.pads[2] + self.pads[5], wshape[4],
                                             self.strides[2],
                                             self.dilations[2])
            shape = [xshape[0], outc, od, ow, oh]
            if volume(self.output_shape) != 0:
                shape[2:] = self.output_shape
        if len(xshape) == 4:
            ow = _convtranspose_output_shape(xshape[2], self.output_padding[0], self.pads[0] + self.pads[2], wshape[2],
                                             self.strides[0],
                                             self.dilations[0])
            oh = _convtranspose_output_shape(xshape[3], self.output_padding[1], self.pads[1] + self.pads[3], wshape[3],
                                             self.strides[1],
                                             self.dilations[1])
            shape = [xshape[0], outc, ow, oh]
            if volume(self.output_shape) != 0:
                shape[2:] = self.output_shape
        elif len(xshape) == 3:
            ow = _convtranspose_output_shape(xshape[2], self.output_padding[0], self.pads[0] + self.pads[1], wshape[2],
                                             self.strides[0],
                                             self.dilations[0])
            shape = [xshape[0], outc, ow]
            if volume(self.output_shape) != 0:
                shape[2] = self.output_shape[0]
        outtensors[0].update_shape(shape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        macs = 0
        if len(outtensors) == 1:
            if len(intensors) == 3 or len(intensors) == 2:
                kernel_shape = intensors[1].get_shape()
                outvol = volume(outtensors[0].get_shape())
                reduce_shape = kernel_shape[1:]
                reduce_vol = volume(reduce_shape)
                macs += outvol * reduce_vol * MUL_MACS
                if len(intensors) > 2:
                    macs += (outvol * ADD_MACS)
        return [macs, 0]


@NODE_REGISTRY.register()
class ReshapeNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        srcshape = intensors[0].get_shape()
        if not is_valid_ndarray(intensors[1].get_numpy()):
            outtensors[0].update_shape([1, ])
            outtensors[0].update_dtype(intensors[0].dtype)
            return
        shape = intensors[1].get_numpy()
        newshape = []
        for i in range(len(shape)):
            if shape[i] == 0:
                newshape.append(int(srcshape[i]))
            else:
                newshape.append(int(shape[i]))
        sum = volume(newshape)
        raw = volume(srcshape)
        if sum < 0:
            remain = raw // -sum
            for i, val in enumerate(newshape):
                if val == -1:
                    newshape[i] = remain
                    break
        assert raw == volume(newshape)
        outtensors[0].update_shape(newshape)
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        shape = []
        xtensor = intensors[0].get_numpy()
        stensor = intensors[1].get_numpy()
        for i,v in enumerate(stensor):
            if v == 0:
                shape.append(xtensor.shape[i])
            else:
                shape.append(v)
        ret = xtensor.reshape(shape)
        outtensors[0].update_tensor(ret)


@NODE_REGISTRY.register()
class GatherElementsNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        x = intensors[0].get_numpy()
        indice = intensors[1].get_numpy()
        outtensor = numpy.zeros_like(indice)
        for i in numpy.ndindex(outtensor.shape):
            idx = list(i)
            idx[self.axis] = indice[i]
            outtensor[i] = x[tuple(idx)]
        outtensors[0].update_tensor(outtensor)


@NODE_REGISTRY.register()
class GRUNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        seq_len = xshape[0]
        batch = xshape[1]
        num_dir = wshape[0]
        h_len = wshape[1] // 3
        outtensors[0].update_shape([seq_len, num_dir, batch, h_len])
        outtensors[0].update_dtype(intensors[0].dtype)
        outtensors[1].update_shape([num_dir, batch, h_len])
        outtensors[1].update_dtype(intensors[0].dtype)

    def profile(self, intensors: List[Tensor], outtensors: List[Tensor]):
        xshape = intensors[0].get_shape()
        wshape = intensors[1].get_shape()
        rshape = intensors[2].get_shape()
        bshape = intensors[3].get_shape()
        batch = xshape[1]
        macs = volume(wshape) + volume(rshape) + volume(bshape) * ADD_MACS
        macs *= batch
        return [macs, 0]


@NODE_REGISTRY.register()
class ConstantOfShapeNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('value', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(list(intensors[0].get_numpy()))
        if self.value is None:
            outtensors[0].update_dtype(numpy.float32)
        else:
            outtensors[0].update_dtype(self.value.dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        arr = numpy.zeros(intensors[0].get_numpy(), dtype=self.value.dtype)
        if self.value is not None and len(self.value) == 1:
            arr.fill(self.value[0])
        outtensors[0].update_tensor(arr)


@NODE_REGISTRY.register()
class CastNode(Node):
    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_shape(intensors[0].get_shape())
        outtensors[0].update_dtype(onnxdtype2npdtype(self.to))

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        outtensors[0].update_tensor(intensors[0].get_numpy().astype(onnxdtype2npdtype(self.to)))


@NODE_REGISTRY.register()
class SplitNode(Node):
    def __init__(self, nodeproto):
        super().__init__(nodeproto)
        self.add_default_value('axis', None)
        self.add_default_value('split', None)

    def shape_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        end = 0
        inshape = intensors[0].get_shape()
        if self.split is None:
            if len(intensors) == 2:
                split = intensors[1].get_numpy()
            else:
                split = [inshape[self.axis] // 2] * 2
        else:
            split = self.split
        axis = _axes_neg2pos(len(inshape), [self.axis])[0]
        shapes = []
        for v in split:
            shape = []
            for i in range(len(inshape)):
                if i == axis:
                    if end + v < inshape[axis]:
                        shape.append(v)
                    else:
                        shape.append(inshape[axis] - end)
                else:
                    shape.append(inshape[i])
            end += v
            shapes.append(shape)
        for i in range(len(shapes)):
            outtensors[i].update_shape(shapes[i])
            outtensors[i].update_dtype(intensors[0].dtype)

    def value_infer(self, intensors: List[Tensor], outtensors: List[Tensor]):
        splitpos = []
        end = 0
        inshape = intensors[0].get_shape()
        if self.split is None:
            if len(intensors) == 2:
                split = intensors[1].get_numpy()
            else:
                split = [inshape[self.axis] // 2]
        else:
            split = self.split

        axis = _axes_neg2pos(len(inshape), [self.axis])[0]
        for v in split:
            if end + v >= inshape[axis]:
                break
            splitpos.append(end + v)
            end += v
        ret = numpy.split(intensors[0].get_numpy(), splitpos, axis)
        for i,t in enumerate(ret):
            outtensors[i].update_tensor(t)


def create_node(n: onnx.NodeProto):
    node_class = NODE_REGISTRY.get(n.op_type + 'Node')
    if node_class != None:
        instance = node_class(n)
        return instance
    warnings.warn(f'node {n.op_type} is not registed for profiling, return 0 Macs and 0 params as default. '
                  f'Use NODEPROFILER_REGISTRY to register your profiler for this node.')
    return Node(n)
